# @Version: $Id: Alan.pm 121 2017-04-22 07:56:24Z vlbi $
# Module for computing Allan variance
package Alan;
use Exporter;
use Carp;

@ISA = qw(Exporter);

#@EXPORT = qw(Mjd Doy Soy Sod Print_fmt1 Print_fmt2 Print_fmt3 
#	     Print_fmt4 Print_fmt10 Str_fmt3 Set_mjd2Time 
#	     Set_time_strydhms Set_time_strymdhms Set_time_ydhms, 
#            Set_time_Max);
#  Max time = 3000/1/1 

BEGIN{
}


sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this->{xfact} = 1.0;
    $this->{yfact} = 1.0;
    $this->{xerr} = 1.e-5; # Allowable xdata range for the epoch.
    $this;
}

sub Set_Xallow($$){
    my $pkg =shift;
    my $xerr =shift;
    $pkg->{xerr} = $xerr;
}

# X/Yfact are factor to multiply for alan-var.
sub Set_Xfact($$){
    my $pkg =shift;
    my $fact =shift;
    $pkg->{xfact} = $fact;
}
sub Set_Yfact($$){
    my $pkg =shift;
    my $fact =shift;
    $pkg->{yfact} = $fact;
}

sub Set_Arr($\@\@){
    my $pkg =shift;
    my $xarr =shift;
    my $yarr =shift;
    $pkg->{Xarr} = $xarr;
    $pkg->{Yarr} = $yarr;
    my $ndat = @{$yarr};
#    $ndat = @{$xarr} if(@{$yarr} < $ndat);
    $pkg->{ndat} = $ndat;
    
    my $m1=0;
    my $cnt=0;
    for($i=0; $i<$ndat; $i++){ 
	next unless(defined($$yarr[$i]));
	$m1 += $$yarr[$i];  $cnt++;  }
    my $yavg = $m1/$cnt;
    my $sgn = $yavg > 0 ? 1 : -1;
    my $yord = int(log($yavg * $sgn)/log(10));
    my $yfact_intnl = 10**$yord;
    $pkg->{yfact_intnl} = $yfact_intnl;

    return $cnt;
}

# Computing Alan Variance with Xarr and Yarr
sub CompAlanStd($$;$){
    my $pkg =shift;
    my $intvl =shift;
    my $xerr =shift;
    $xerr = $pkg->{xerr} unless(defined($xerr));
    my $xarr = $pkg->{Xarr};
    my $yarr = $pkg->{Yarr};

    my $m2=0;
    my $cnt=0;
    my $ndat = @$yarr;
    unless(defined($$xarr[$ndat-1])){
	print STDERR "Err: xarr[$ndat - 1] is not defined.\n";
	return (0,0);
    }
    if( $$xarr[$ndat-1] <= 2*$intvl ){
	print STDERR "E Alan::CompAanStd(): xarr_end($$xarr[$ndat-1]) =< intvl($intvl) \n";
	return (-1, -1);
    }
    my ($i1, $i2) = (0,0);
    for( my $i=0; $i<$ndat; $i++){
	my $x0 = ${$xarr}[$i];
	my $x_exp1 = $x0 + $intvl;
	my $x_exp2 = $x0 + $intvl*2.0;
	($i1, $i2) = ($i,$i);
	my $j=$i;
	while( ($j < $ndat) && (($i1 == $i) || ($i2 == $i)) ){
	    if( abs($x_exp1 - ${$xarr}[$j]) < $xerr){
		$i1 = $j; $j++;
	    }elsif( abs($x_exp2 - ${$xarr}[$j]) < $xerr){
		$i2 = $j; $j++;
	    }else{ $j++; next;}
	}
	if(($i1 != $i) && ($i2 != $i)){
	    my $y1 = ${$yarr}[$i] + ${$yarr}[$i2] - 2.0*${$yarr}[$i1];
	    $m2 += ($y1 * $y1);
	    $cnt++;
	    $i=$i2-1;
	}
	
    }

    my $fact = $pkg->{yfact}/$pkg->{xfact};
    if($cnt <= 0){
	print STDERR "Err Alan::CompAlanStd(): not matching interval($intvl) data. \n";
	return(0,0);
    }
    my $alvar = $m2/$cnt;
    my $alstd = sqrt($alvar)*0.5 *$fact/$intvl;;
    return ($alstd, $cnt);
}

# Computing Alan Variance with Yarr assuming constnt intterval
sub CompAlanStd2($$){
    my $pkg =shift;
    my $intvl =shift;
    my $yarr = $pkg->{Yarr};

    my $ndat = @$yarr;
    my $xunit = $pkg->{xfact};    
    my $yunit = $pkg->{yfact};

    if($ndat <= 2*$intvl ){
	print STDERR "E Alan::CompAanStd(): ndat($ndat) =< intvl($intvl) \n";
	return (-1, -1);
    }
    my ($i1, $i2) = (0,0);
    my $m2=0;
    my $cnt=0;
	
    my $yfact_intnl =$pkg->{yfact_intnl};
    for( my $i=0; $i + $intvl*2 <$ndat; $i += $intvl*2){
	$y1= ${$yarr}[$i+2*$intvl] + ${$yarr}[$i] - 2.0* ${$yarr}[$i+$intvl];  
	$y1 /= $yfact_intnl;
	$m2 += $y1*$y1;
	$cnt++;
    }
    my $y2avg = $m2/$cnt;

    my $fact = $yunit/$xunit;
    my $alstd = sqrt($y2avg / 2.0) *$fact/$intvl * $yfact_intnl;
    return ($alstd, $cnt);
}

# Computing Modified Alan Variance with xarr and yarr data
sub CompMAlanStd($$;$){
    my $pkg =shift;
    my $intvl =shift;
    my $xerr =shift;
    $xerr = $pkg->{xerr} unless(defined($xerr));
    my $xarr = $pkg->{Xarr};
    my $yarr = $pkg->{Yarr};

    my $m2=0;
    my $cnt=0;
    my $ndat = @$xarr;
    my ($i1, $i2) = (0,0);
    for( my $i=0; $i<$ndat; $i++){
	my $x0 = ${$xarr}[$i];
	my $x_exp1 = $x0 + $intvl;
	my $x_exp2 = $x0 + $intvl*2.0;
	($i1, $i2) = ($i,$i);
	my $j=$i;
	while( ($j < $ndat) && (($i1 == $i) || ($i2 == $i)) ){
	    if( abs($x_exp1 - ${$xarr}[$j]) < $xerr){
		$i1 = $j; $j++;
	    }elsif( abs($x_exp2 - ${$xarr}[$j]) < $xerr){
		$i2 = $j;$j++;
	    }else{ $j++; next;}
	}
	if(($i1 != $i) && ($i2 != $i)){
	    my $y1 = ${$yarr}[$i] + ${$yarr}[$i2] - 2.0*${$yarr}[$i1];
	    $m2 += ($y1 * $y1);
	    $cnt++;
#	    $i=$i2-1;
	}
	
    }

    my $fact = $pkg->{yfact}/$pkg->{xfact};

    my $alvar = $m2/$cnt;
    my $alstd = sqrt($alvar / 2) *$fact/$intvl;;
    return ( $alstd, $cnt);
}

# Computing Modified Alan Variance with only yarr data assuming constant interval
sub CompMAlanStd2($$){
    my $pkg =shift;
    my $intvl =shift;
    my $yarr = $pkg->{Yarr};

    my $xunit = $pkg->{xfact};    
    my $yunit = $pkg->{yfact};

    my $m2=0;
    my $cnt=0;
    my $ndat = @$yarr;
    my $yfact_intnl =$pkg->{yfact_intnl};
    
    for( my $i=0; $i+$intvl*2<$ndat; $i ++){
	$y1= ${$yarr}[$i+2*$intvl] + ${$yarr}[$i] - 2.0* ${$yarr}[$i+$intvl];  
	$y1 /= $yfact_intnl;
	$m2 += $y1*$y1;
	$cnt++;
    }
    my $y2avg = $m2/$cnt;

    my $fact = $yunit/$xunit;
    my $alstd = sqrt($y2avg/2.0) *$fact/$intvl * $yfact_intnl;
    return ($alstd, $cnt);
}


sub Intpl($$){
    my $pkg = shift;
    my $x = shift;
    my $xarr = $pkg->{Xarr};
    my $yarr = $pkg->{Yarr};
    my $ndat = $pkg->{ndat};

    my $i=0;
    $i++ while($i<$ndat &&  $x > $$xarr[$i]);
    if( $i >= $ndat){
	return (0,-1);
    }else{
	my $y = ($$yarr[$i] - $$yarr[$i-1])/ ($$xarr[$i] - $$xarr[$i-1]) *
	    ($x - $$xarr[$i]) + $$yarr[$i];
	return ($y, $i-1);
    }

}
# Computing Interpolated Alan Variance
# maxintvl is maximum interval use for interpolation. A sample data is skipped, 
# if adjucent data interval is larger than this. 
sub CompIAlanStd($$$){
    my $pkg =shift;
    my $intvl =shift;
    my $maxintvl =shift;
    my $xarr = $pkg->{Xarr};
    my $yarr = $pkg->{Yarr};

    my $m2=0;
    my $cnt=0;
    my $ndat = @$xarr;
    my ($i1, $i2) = (0,0);
    for( my $i=0; $i<$ndat; $i++){
	my $x0 = ${$xarr}[$i];
	my $x_exp1 = $x0 + $intvl;
	my $x_exp2 = $x0 + $intvl*2.0;
	my ($yv1, $i1) = $pkg->Intpl($x_exp1);
	my ($yv2, $i2) = $pkg->Intpl($x_exp2);
	next if( $i1 <0 || $i2 < 0);
	my $dt1 = abs($x_exp1 - ${$xarr}[$i1]);
	my $dt2 = abs($x_exp2 - ${$xarr}[$i2]);
	next if(($dt1 > $maxintvl) || ($dt2 > $maxintvl));

	my $y1 = ${$yarr}[$i] + $yv2 - 2.0*$yv1;
	$m2 += ($y1 * $y1);
	$cnt++;
    }
	
    return ( 0, 0) if($cnt==0);

    my $fact = $pkg->{yfact}/$pkg->{xfact};

    my $alvar = $m2/$cnt;
    my $alstd = sqrt($alvar/2.0) *$fact/$intvl;;
    return ( $alstd, $cnt);
}


# Statisitical diff of Xdata
sub XDiff_Stat($$){
    my $pkg=shift;    
    my $scl=shift;
    my $xarr = $pkg->{Xarr};
    my $ndat = @{$xarr};
    my $tdiff;
    my %dst;
    for(my $i=0; $i<$ndat-1; $i++){ 
	my $tdiff = $$xarr[$i+1] - $$xarr[$i]; 
	my $rdiff = int($tdiff / $scl + 0.5) * $scl;
	my @lst = keys(%dst);
	if(grep(/$rdiff/,@lst)){$dst{$rdiff} += 1;}
	else{ $dst{$rdiff}=1;}
    }
    
    return %dst;
    
    
}


1;

