# /**
#  * @file  Blistl.pm
#  *
#  * @author M.Sekido (NICT Kashima.)
#  * @date $Date: 2016-11-01 10:18:22 +0900 (Tue, 01 Nov 2016) $
#  * @version $Id: Blist.pm 112 2016-11-01 01:18:22Z vlbi $
#  * 
#  * Copyright (C) 2007 NICT Kashima. All rights reserved.
#  */

package Blist;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw();

#use PGPLOT;

BEGIN {
    %colno = (
	'OBS' =>1,
	'BAND' =>2,
	'SRC' =>3,
	'BASE' =>4,
	'TM' =>5,
	'PAMP'=>6,
	'GD' =>7,  # Changed from SD to GD: Group delay 2016.6.24
	'PD' =>8,  # PD is no more precise delay but Phase delay 2016.6.24
	'PRAT'=>9,
	'SNR' =>10,
	'QCODE'=>11,
	'AMBG'=>12,
	'Tint' =>13,
	'ECSD' =>14,
	'EGPD' =>15,
	'ERAT' =>16,
	'AZ1' =>17,
	'AZ2' =>18,
	'EL1' =>19,
	'EL2' =>20
	      );
 # old Blist format
#     %colno = (
# 	      'OBS' =>1,
# 	      'BAND' =>2,
# 	      'SRC' =>3,
# 	      'BASE' =>4,
# 	      'TM' =>5,
# 	      'PAMP'=>6,
# 	      'GD' =>7,  # Changed from SD to GD: Group delay 2016.6.24
# 	      'PD' =>8,  # PD is no more precise delay but Phase delay 2016.6.24
# 	      'PRAT'=>9,
# 	      'SNR' =>10,
# 	      'AMBG'=>12,
# 	      'Tint' =>13,
# 	      'ECSD' =>14,
# 	      'EGPD' =>15,
# 	      'ERAT' =>16
# 	      );
}

use Maxmin;

sub new ($){
    my $pkg = shift;
    my $this = { };
    bless $this;
    $this;
}

sub ColNo($$){
    my $pkg = shift;
    my $key = shift;
    return $colno{$key};
}
sub Read($$){
    my $pkg = shift;
    my $fname = shift;
    unless( open(IN,$fname) ){
	print STDERR "Error: in opening the data file ($fname)\n";
	return -1;
    }

#    my (@tm, @obs,@src,@pdel,@cdel,@pamp,@snr,
    my (@tm, @obs,@src,@pdel,@gdel,@pamp,@snr,
	@prat,@ambg,@tint,@ecsd,@egpd,@erat, @serno,
	@az1, @az2, @el1, @el2);
    my ($base, $band);
    use Time_lib;

    my $cnt =0;
    while(<IN>){
	next if(/OBS  BAND  SRC  BASE/);
	next if(/^\#/);
	my @terms = split;
	next if( @terms < 17);
	my $icol =$colno{"TM"};
	if($terms[$icol] =~/(\d{4})(\d{3})(\d\d)(\d\d)(\d\d)/ ||
	   $terms[$icol] =~/(\d{4}).(\d{3}).(\d\d):(\d\d):(\d\d)/){
	    my @hms = ($1,$2,$3,$4,$5);
	    my $tm = new Time_lib;
	    $tm->Set_time_ydhms(@hms);
	    $tm[$cnt] = $tm;
	}else{
	    print "E Timeformat is unexpected: $terms[$icol] \n";
	    next;
	}
	$serno[$cnt] = $cnt;
	$obs[$cnt] = $terms[$colno{"OBS"}];
	$src[$cnt] = $terms[$colno{"SRC"}];
	$base = $terms[$colno{"BASE"}];
	$band = $terms[$colno{"BAND"}];
	$pdel[$cnt] = $terms[$colno{"PD"}];
#	$cdel[$cnt] = $terms[$colno{"SD"}];
	$gdel[$cnt] = $terms[$colno{"GD"}];
	$pamp[$cnt] = $terms[$colno{"PAMP"}];
	$snr[$cnt] = $terms[$colno{"SNR"}];
	$prat[$cnt] = $terms[$colno{"PRAT"}];
	$ambg[$cnt] = $terms[$colno{"AMBG"}];
	$tint[$cnt] = $terms[$colno{"Tint"}];
	$ecsd[$cnt] = $terms[$colno{"ECSD"}];
	$egpd[$cnt] = $terms[$colno{"EGPD"}];
	$erat[$cnt] = $terms[$colno{"ERAT"}];
	$az1[$cnt] = $terms[$colno{"AZ1"}] if(defined($terms[$colno{"AZ1"}]));
	$az2[$cnt] = $terms[$colno{"AZ2"}] if(defined($terms[$colno{"AZ2"}]));
	$el1[$cnt] = $terms[$colno{"EL1"}] if(defined($terms[$colno{"EL1"}]));
	$el2[$cnt] = $terms[$colno{"EL2"}] if(defined($terms[$colno{"EL2"}]));
	$cnt++;
    }
    ## Sorting Serno, by Time data
    my @serno2 = sort {$tm[int($a)] <=> $tm[int($b)] } @serno;


    $pkg->{SN} = \@serno;
    $pkg->{SN2} = \@serno2;
    $pkg->{TM} = \@tm;
    $pkg->{OBS} = \@obs;
    $pkg->{SRC} = \@src;
    $pkg->{PD} = \@pdel;
#    $pkg->{SD} = \@cdel;
    $pkg->{GD} = \@gdel;
    $pkg->{PAMP} = \@pamp;
    $pkg->{SNR} = \@snr;
    $pkg->{PRAT} = \@prat;
    $pkg->{AMBG} = \@ambg;
    $pkg->{Tint} = \@tint;
    $pkg->{ESCD} = \@ecsd;
    $pkg->{EGPD} = \@egpd;
    $pkg->{ERAT} = \@erat;
    $pkg->{BASE} = $base;
    $pkg->{BAND} = $band;
    $pkg->{AZ1} = \@az1;
    $pkg->{AZ2} = \@az2;
    $pkg->{EL1} = \@el1;
    $pkg->{EL2} = \@el2;
    return $cnt;
}

## Return Serial Number in time order
sub SerNo($){
    $pkg = shift;
    return @{$pkg->{SN2}};
}

sub Tarr($;$){
    $pkg = shift;
    $ser = shift;
    my $tm = $pkg->{TM};
    my @tmarr;
    if(defined($ser)){
	@tmarr= map($$tm[$_],@{$ser});
    }else{
	@tmarr= @{$tm};
    }
    return @tmarr;
}

sub Soyarr($;$){
    $pkg = shift;
    $ser = shift;
    my $tm = $pkg->{TM};
    my @tmarr;

    if(defined($ser)){
	@tmarr= map($$tm[$_],@{$ser});
    }else{
	@tmarr= @{$tm};
    }
    @soy = map($_->Soy(), @tmarr);
    return @soy;
}

sub Base($){
    my $pkg = shift;
    return $pkg->{BASE};
}
sub Band($){
    my $pkg = shift;
    return $pkg->{BAND};
}

sub Datarr($$;$){
    my $pkg = shift;
    my $key = shift;
    my $ser = shift;
    my $arr = $pkg->{$key};
    if(defined($ser)){
	return map($$arr[$_],@$ser);
    }else{
	return @{$arr};	
    }

}

## Return the data of that source
sub SRC_dat($$$){
    my ($pkg,$src,$key)=@_;
    my $darr = $pkg->{$key};
    my $srcarr = $pkg->{SRC};
    if($srcarr ==0){
	print STDERR "No scan for source($src) were found\n";
	return ();
    }
    my $arr;  $#arr=-1;
    my $cnt=0;
    for(my $i=0; $i<@$darr; $i++){
#	print STDERR "$src  $$srcarr[$i] \n";
	if($$srcarr[$i] eq $src){
	    if($key eq "SN"){
		$arr[$cnt++] = $i;
	    }else{
		$arr[$cnt++] = $$darr[$i];
	    }
	    printf STDERR "$src Matched %d\n", $#arr +1;
	} 
    }
    
    return @arr;
}

1;
