## DBHD3 output reading module
## coded by M.Sekido 2014/9/4
# @Version: $Id: Time_lib.pm 60 2015-08-12 04:44:20Z sekido $
package DBext;

use Time_lib;
BEGIN{
    %TOC2_FLAG = (0x01 => "SEC TAG",  0x02 => "DEL OBSV",
		  0x04 => "DELSIGMA", 0x08 => "RAT OBSV", 
		  0x10 => "RATSIGMA", 0x20 => "DELRESID",
		  0x40 => "RATRESID", 0x80 => "UTC TAG", 
		  0x100 =>"BASELINE" 
	);
}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

sub Read($$){
    my $pkg = shift;
    my $fname = shift;
    if( $pkg->Read_Toc1($fname) < 0){
	print STDERR "E Failed to read TOC1 \n";
	return -1;
    }
    if( $pkg->Read_Toc2($fname) < 0){
	print STDERR "E Failed to read TOC2 \n";
	return -1;
    }
    
}

sub Read_Toc1($$){
    my $pkg = shift;
    my $fname = shift;
    unless(open(IN,$fname)){
	print STDERR "E DBext::Read_Toc1(): Failed to open infile($fname)\n";
	return -1;
    }

    my ($nstn, $nstn2);
    

    my $buf = <IN>;
    while( !eof(IN) && $buf !~/record type\s+1/){  $buf = <IN>; }
    unless( $buf =~/record type\s+1/ ){
	print "E failed to find TOC1\n";
	return -1;
    }

    while( !eof(IN) && $buf !~/SITNAMES/){  $buf = <IN>; }
    if( eof(IN)){
	print STDERR "E SITNAMES Entory not found.\n";
	return -1;
    }else{	$nstn = $pkg->Get_StnName(IN);    }

    seek(IN,0, 0);  $buf = <IN>;
    $buf = <IN>  while( !eof(IN) && $buf !~/SITERECS/);
    if( eof(IN)){
	print STDERR "E SITERECS Entory not found.\n";
#	return -1;
    }else{	$nstn2 = $pkg->Get_StnXYZ(IN); }

    seek(IN,0, 0);  $buf = <IN>;
    $buf = <IN>  while( !eof(IN) && $buf !~/STRNAMES/);
    if( eof(IN)){
	print STDERR "E STRNAMES Entory not found.\n";
	return -1;
    }else{    $nstn2 = $pkg->Get_StrName(IN); }

    seek(IN,0, 0);  $buf = <IN>;
    my @nobs;
    $buf = <IN>  while( !eof(IN) && $buf !~/NUMB OBS/);
    if( eof(IN)){
	print STDERR "E Nobs Entory not found.\n";
	$nobs[0] = 0;
    }else{
	my $ret = $pkg->Get_1val(IN); 
	if(defined($ret)){$nobs[0]=$ret;}
	else{ print STDERR "E DBext::Read(): failed to get 'NUMB OBS'\n"; return -1;} 
    }
    $buf = <IN>  while( !eof(IN) && $buf !~/NUM4 OBS/);
    if( eof(IN)){
	print STDERR "E Nobs2 Entory not found.\n";
	$nobs[1] = 0;
    }else{   
	my  $ret = $pkg->Get_1val(IN); 
	if(defined($ret)){ $nobs[1] = $ret;}
	else{ print STDERR "E DBext::Read(): failed to get 'NUM4 OBS'\n"; return -1;} 
    }
    $pkg->{nobs} = \@nobs;

    seek(IN,0, 0);  $buf = <IN>;
    $buf = <IN>  while( !eof(IN) && $buf !~/ERROR K.*\s+\(\s+(.+)\s+\)/);
    my $dumy = $1;
    my ($ndr, $nbase, $nd)=split(/\s+/,$dumy);
    
    $pkg->{nbase} = $nbase;
    if( eof(IN)){
	print STDERR "E 'ERROR K' (Additional Error) Entory not found.\n";
    }else{ 
	my @base_err;
	for(my $i=0; $i<$nbase; $i++){
	    my @arr = $pkg->Get_nitem(IN, $ndr,1);    
	    $base_err[$i] = \@arr;
	}
	$pkg->{error_k} = \@base_err;
    }


    seek(IN,0, 0);  $buf = <IN>;
    $buf = <IN>  while( !eof(IN) && $buf !~/ERROR BL.*\s+\(\s+(.+)\s+\)/);
    $dumy = $1;
    ($nd, $ndr, $nbase)=split(/\s+/,$dumy);
    
    if( eof(IN)){
	print STDERR "E 'ERROR BL' (Additional Error) Entory not found.\n";
    }else{ 
	my @base_err;
	for(my $i=0; $i<$nbase; $i++){
	    my @arr = $pkg->Get_nitem(IN, $ndr,1);    
	    $base_name[$i] = join(":",@arr);
	}
	$pkg->{error_bl} = \@base_name;
    }
    
    close(IN);
    return 0;
}

sub Nbase($){
    $pkg = shift;
    $pkg->{nbase};

}

sub Read_Toc2($$){
    my $pkg = shift;
    my $fname = shift;
    unless(open(IN,$fname)){
	print STDERR "E DBext::Read_Toc2(): Failed to open infile($fname)\n";
	return -1;
    }

    my ($nstn, $nstn2);
    

    my $buf = <IN>;
    my $cnt=0;
    my $bit_flag =0;
    my $max_flag = 2**9 -1;
    my @dbarr;
    my $itoc=0;
    my $new_itoc=0;
    my $toctype = 0;
    while( !eof(IN) ){
	if( $buf =~ /(\d+)\s+of record type\s+(\d)/){
	    $new_itoc = $1;
	    $toctype = $2;
	}
	next if( $toctype != 2 || $buf !~/\S+/ );
	
	if( $itoc != $new_itoc){
	    if( $itoc !=0 && $bit_flag != $max_flag ){
		printf "W TOC2 [%d]th Ocrnc was not enough %d/%d\n", 
		$itoc, $bit_flag, $max_flag;
		foreach $k (keys(%TOC2_FLAG)){
		    printf "W %s was missing \n",$TOC2_FLAG{$k}
		    unless ($bit_flag & $k);
		}
	    }
	    $bit_flag =0;
	    my %db;
	    $itoc = $new_itoc;
	    $dbarr[$itoc] = \%db;
	    print STDERR "." if($itoc % 100 ==0 );
	    next;
	}
	
	
	foreach $k (keys(%TOC2_FLAG)){
	    my $val= $TOC2_FLAG{$k};
	    next unless( $buf =~/$val/ );
#	    print "DBG: $val \n";

	    $bit_flag |= $k;
	    if($k == 0x01){
		$dbarr[$itoc]->{sec} = $pkg->Get_1val(IN);
	    }elsif($k == 0x02){
		$dbarr[$itoc]->{delobs} = $pkg->Get_2to1val(IN);
	    }elsif($k == 0x04){
		$dbarr[$itoc]->{delsigma} = $pkg->Get_1val(IN);
	    }elsif($k == 0x08){
		$dbarr[$itoc]->{ratobs} = $pkg->Get_1val(IN);
	    }elsif($k == 0x10){
		$dbarr[$itoc]->{ratsigma} = $pkg->Get_1val(IN);
	    }elsif($k == 0x20){
		$dbarr[$itoc]->{delrsd} = $pkg->Get_1val(IN);
	    }elsif($k == 0x40){
		$dbarr[$itoc]->{ratrsd} = $pkg->Get_1val(IN);
	    }elsif($k == 0x80){
		my @ymdhm = $pkg->Get_nitem(IN,5);
		$ymdhm[0] += $ymdhm[0] > 50 ? 1900 : 2000;
		my $sec = $dbarr[$itoc]->{sec};
		my $t1 = new Time_lib;
		$t1->Set_time_ymd(@ymdhm,$sec);
		$dbarr[$itoc]->{utctag} = $t1;
	    }elsif($k == 0x100){
		my @bl= $pkg->Get_list(IN,2);
		$dbarr[$itoc]->{baseline} = join(":",@bl);
	    }
	} # foreach end
	
    }continue{
	$buf =<IN>;
    }
    $pkg->{toc2db} = \@dbarr;
    close(IN);
    return 0;
}

# Return 'NUMB OBS' and 'NUM4 OBS'
sub Nobs($){
    $pkg = shift;
    return @{$pkg -> {nobs}};
}

# IN, Numdata, D/E replace flag
sub Get_nitem($$$;$){
    $pkg = shift;
    my $in = shift;
    my $ndat = shift;
    my $DE_flag = shift; # Flag to replace D/Exp
    my $p = caller;
    my $I = $p . "::" . $in;
    my @arr =();
    my $indat = $ndat;
    while( ($indat > 0) && (!eof($I)) ){
	my $buf = <$I>;
	next if ( $buf !~/\S+/);
	$buf =~s/D/E/g	if(defined($DE_flag) && $DE_flag ==1);
	$buf =~/^\s*/;
	my @terms = split(/\s+/,$');   #');
	my $nitm = @terms;
	push(@arr, @terms);
	$indat -= $nitm;
    }
    if($indat > 0){
	printf "E DBext::Get_nitem(): lack of data %d/%d\n", $indat, $ndat;
    }
    return @arr;

}
## 
sub Get_list($$$){
    $pkg = shift;
    my $in = shift;
    my $nl = shift;
    
    my $p = caller;
    my $I = $p. "::" . $in;
    my @bl;
    for(my $i=0; $i<$nl; $i++){
	my $buf = <$I>;
	unless($buf =~/(\S+)/){
	    print "E Unexpected empty in reading baseline.\n";
	}
	$bl[$i] = $1;
    }
    return @bl;

}

sub Err_baseline($$){
    my $pkg = shift;
    my $ibase = shift;
    my $base_name = $pkg->{error_bl};
    if(defined($base_name)){
	return $$base_name[$ibase];
    }else{
	return undef;
    }
}

# Return Delay, rate add_sigma
 sub ErrorK($$){
    my $pkg = shift;
    my $ibase = shift;
    my $base_err = $pkg->{error_k};
    if(defined($base_err)){
	return @{$$base_err[$ibase]};
    }else{
	return undef;
    }
}

# REturn in [usec]
sub Toc2_DelObs($$)
{
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{delobs};
}

# Return in [s]
sub Toc2_DelRsd($$)
{
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{delrsd};
}

# Return in [s]
sub Toc2_RatRsd($$)
{
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{ratrsd};
}

# Return in [s]
sub Toc2_DelSig($$)
{
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{delsigma};
}

# Return in [s/s]
sub Toc2_RatObs($$)
{
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{ratobs};
}
# Return in [s/s]
sub Toc2_RatSig($$)
{
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{ratsigma};
}

sub Toc2_Baseline($$){
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    return $$toc2db[$itoc]->{baseline};
}

# Return Time_lib object reference 
sub Toc2_UTC($$){
    my $pkg = shift;
    my $itoc = shift;
    my $toc2db = $pkg->{toc2db};
    my $t1 = $$toc2db[$itoc]->{utctag};
    return $t1;
}

sub Get_1val($$){
    my ($pkg, $in) = @_;
    my $p = caller;
    my $I = $p. "::" . $in;
    my $buf = <$I>;
    my $ptn = "[-\+\\dD+.]+";
    my $i=0;
    my $val;
    for( $i=0; $i<3; $i++){
	if ($buf =~ /($ptn)/){
	    $val=$1;  $val =~s/D/e/;   last };
	$buf = <$I> ;
    } 
    if($i == 3){
	return undef;
    }else{
	return $val;
    }
}
sub Get_5val($$){
    my ($pkg,$in) = @_;
    my $p = caller;
    my $I = $p. "::" . $in;
    my $buf = <$I>;
    $buf =~/\s+(.*)/;
    my @ymdhm = split(/\s+/);
    return @ymdhm;
}


# Retrun in [usec] 
sub Get_2to1val($$){
    my ($pkg,$in) = @_;
    my $p = caller;
    my $I = $p . "::" . $in;
    my $buf = <$I>;
    my $ptn = "[\\dD+-.]+";
    $buf = <$I>   while($buf =~/($ptn)\s+($ptn)/);
    my $val1=$1;    my $val2=$2;
    $val1 =~s/D/e/; $val2 =~s/D/e/;
    my $val = $val1 + $val2;
    return $val; # [usec]
}

sub Get_StnXYZ($$){
    my ($pkg,$in) = @_;
    my $p = caller;
    my $I = $p. "::" . $in;
    my $stndb = $pkg->{stndb};
    my %stndb;
    $stndb = \%stndb unless(defined($stndb));
    my (@stnxyz);
    my $cnt =0;
    while( <$I>){
	next if(/^\*/);
	last if(/Lcode/);
	next unless(/\s*(\S.+)/);
	my $xyz = $1;
	$xyz =~s/D/e/g;
	my @terms = split(/\s+/,$xyz);
	$stnxyz[$cnt] = \@terms;
	$cnt++;
    }

    $stndb->{stnxyz} = \@stnxyz;
    $pkg->{stndb} = $stndb;
    if(  $pkg->{nstn} != $cnt){
	printf STDERR "W DBext::Get_stnxyz(): Nstn(%d) != count(%d)\n",
	$pkg->{nstn} ,$cnt;
    };
    return  $cnt;
}

sub StnXYZ($){
    my $pkg=shift;
    $stndb = $pkg->{stndb};
    return @{$stndb->{stnxyz}};
}

sub Nstn($){
    my $pkg= shift;
    return $pkg->{nstn};
}

sub Get_StnName($$)
{
    my ($pkg,$in) = @_;
    my $p = caller;
    my $I = $p. "::" . $in;
    my $stndb = $pkg->{stndb};
    my %stndb;
    $stndb = \%stndb unless(defined($stndb));
    my (@stnname);
    my $cnt =0;
    while( <$I>){
	next if(/^\*/);
	last if(/Lcode/);
	last unless(/(\S+)/);
	$stnname[$cnt] = $1;
	$cnt++;
    }

    $stndb->{stnname} = \@stnname;
    $pkg->{stndb} = $stndb;
    $pkg->{nstn} = $cnt;
    return  $cnt;
}

sub Get_StrName($$)
{
    my ($pkg,$in) = @_;
    my $p = caller;
    my $I = $p. "::" . $in;
    my $strdb = $pkg->{strdb};
    my %strdb;
    $strdb = \%strdb unless(defined($strdb));
    my (@strname);
    my $cnt =0;
    while( <$I>){
	next if(/^\*/);
	last if(/Lcode/ || / Data base/);
	last unless(/^(\S+)/);
	$strname[$cnt] = $1;
	$cnt++;
    }

    $strdb->{strname} = \@strname;
    $pkg->{strdb} = $strdb;
    $pkg->{nstr} = $cnt;
    return  $cnt;
}

sub StnNames($){
    my $pkg=shift;
    $stndb = $pkg->{stndb};
    return @{$stndb->{stnname}};
}

sub StrNames($){
    my $pkg=shift;
    $strdb = $pkg->{strdb};
    return @{$strdb->{strname}};
}

1;
