#  @Version: $Id: FSlog.pm 119 2017-04-11 08:22:12Z vlbi $
#
package FSlog;
use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(Open);

#require "$ENV{HOME}/prog/perl/UTIL/time.pl";
use lib "$ENV{HOME}/prog/perl/module";
use Time_lib;
use POSIX;

BEGIN{

}


sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

# Set labels to be picked up.
# these lavels are readed by Read_unitl() command;
sub Set_labels($$){
    my ($pkg, $llist) = @_;
    my @labels = @{$llist};
    $pkg->{labels} = \@labels;
}

# in format "yyy[/:]ddd hh:mm:ss"
sub Set_start($$){
    my ($pkg, $st) = @_;
    printf STDOUT " Set Start time=%s \n", $st->Str_fmt1(); 
    $pkg->{st} = $st;
}

sub Read($$){
    my ($pkg, $fname) = @_;
    my $in;
    if( $pkg->Open($fname) != 0){
	print STDERR "E FSlog::Read(): failed to open(file=$fname) \n";
	return -1; 
    }
    return $pkg->Read_until();
}

sub Open($$){
    my ($pkg, $fname) = @_;
    my $in;
    unless(open($in,$fname)){
	print STDERR "E File:$fname failed to open logfile: $fname.\n";
	return -1;
    };

#    my $p = caller;
#    $in  = $p . "::" . $in;

    $pkg->{in} = \$in;
    $pkg->Get_Station();
    return 0;
}

sub Stname($){
    my $pkg = shift;
    return $pkg->{Station};
}

## Separating time and data
sub Separate($){
    my $line = shift;
    chomp($line);
    next if($line =~/^\#/ ) ;
    unless($line =~/(\d{4})\.(\d{3})\.(\d\d):(\d\d):([\d\.]+)/){
	print STDERR "E Matching failed: $line\n";
	next;
    }
    my @ydhms;
    $ydhms[0] = $1;
    $ydhms[1] = $2;
    $ydhms[2] = $3;
    $ydhms[3] = $4;
    $ydhms[4] = $5;
    my $dat = $'; # ';
    my $t1 = new Time_lib;
    $t1->Set_time_ydhms(@ydhms);
    return ($t1, $dat);
}

## Read until specfied epoch
sub Read_until($;$){
    my ($pkg, $t2) = @_;
    unless(defined($t2)){
#	print "W: FSlog::Read_until(): time undefined.\n";
	$t2 = new Time_lib;
	$t2->Set_time_now();
#	return -1;
    }
    my $in = ${$pkg->{in}};
    my $st = $pkg->{st};
    if(defined($st)){
	my $line=<$in>;
	my ($t1, $dat) = Separate($line);
	$st = $t1  unless(defined($st));
	while( !eof($in) && ($st > $t1)){
	    $line=<$in>;
	    ($t1, $dat) = Separate($line);
	}
    }
    my @llist = ();
    @llist =@{$pkg->{labels}}   if(defined(@{$pkg->{labels}}));
    my (@tarr,@darr);
    my $cnt=0;
    my $ird=0;
    my $len;
    while( ! eof($in) ){
	my $line=<$in>;
	$len = length($line);
	$ird++;
	print STDERR "." if($ird %5000 == 0);
	my ($t1, $dat) = Separate($line);
#	return ($cnt, $t1, $dat) if( $t2 < $t1  );
	if( $t2 <= $t1  ){
	    seek $in, -$len,1;
	    last ;
	}
	foreach $l (@llist){
	    if($dat =~m!$l!){
		push(@tarr,($t1));
		push(@darr,($dat));
		$cnt++;
	    }
	}
    }    
    $pkg->{tarr} = \@tarr;
    $pkg->{darr} = \@darr;
    return $cnt;
}

sub Get_list($){
    my $pkg = shift;
    my $tarr = $pkg->{tarr};
    my $darr = $pkg->{darr};
    my $cnt = @{$tarr};
    return ($cnt, $tarr, $darr);
}

## te  : Time_lib data of end-time
## trem: term to be find 
sub Find($$$){
    my ($pkg, $term, $te) = @_;
    my $in = ${$pkg->{in}};

    while( ! eof($in) ){
	my $pos = tell $in;
	my $line = <$in>;
	chomp($line);
	my ($t1,$dat) = Separate($line);
	if(defined($te) && $t1 >= $te){
	    seek( $in, $pos, 0);
	    return (undef,$line);
	} 
	return ($t1,$dat) if($dat =~/$term/);
    }

    if( eof($in)){ return (undef,"EOF");}
    else { return 1 ;}
}

sub Get_line($ ){
    my $pkg = shift;

    my $in = ${$pkg->{in}};
    my $line = <$in>;
    return $line;
}

sub Get_Station($){
    my $pkg = shift;
#    my $p = caller;

    $in = ${$pkg->{in}};
#    $in  = $p . "::" . $in;

    while(<$in>){
	if(/location,([\w\s]+),/){
	    $pkg->{Station} = $1;
	}
    }
    seek($in,0,SEEK_SET);
    return 0;
}

## Read Tsys and El informations
sub Read_tsys($$){
    my ($pkg, $fname) = @_;
    my $in;
    unless( open($in, "$fname") ){
	print STDERR "E Read(): Failed to open file $fname \n";
	return -1;
    }
    $pkg->Get_Station();

    my $st = $pkg->{st};
    my $icnt=-1;
    my (@tarr, @az, @el, @tsysx, @tsyss);
    my $line = <$in>;
    while( ! eof($in) ){
	next if($line =~/^\"/ ) ;
	my ($t1, $dat) = Separate($line);	

	# unless($line =~/(\d{4})\.(\d{3})\.(\d\d):(\d\d):([\d\.]+)/){
	#     print STDERR "E Matching failed: $line\n";
	#     next;
	# }
	# my @ydhms;
	# $ydhms[0] = $1;
	# $ydhms[1] = $2;
	# $ydhms[2] = $3;
	# $ydhms[3] = $4;
	# $ydhms[4] = $5;
	# my $dat = $'; # ';
	# my $t1 = new Time_lib;
	# $t1->Set_time_ydhms(@ydhms);

	next if( defined($st) && $t1 < $st);
	if($dat =~/source\=azeluncr,([\d\.]+)d,([\d\.]+)d/){
	    $icnt++;
	    $az[$icnt] = $1;
	    $el[$icnt] = $2;
	    $tarr[$icnt] = $t1;
	}elsif($dat =~m!/tsys/u5,([\d\.\$]+)!){
	    $tsysx[$icnt] = $1;
	    if($1 =~ /\$\$/){
		$tsysx[$icnt] = 10000 ;} 
	}elsif($dat =~m!/tsys/u6,([\d\.\$]+)!){
	    $tsyss[$icnt] = $1;
	    if($1 =~ /\$\$/){
		$tsyss[$icnt] = 10000; } 
	}
    }continue{
	$line = <$in>;
    }
    close($in);

    $pkg->{tsysx} = \@tsysx;
    $pkg->{tsyss} = \@tsyss;
    $pkg->{tarr} = \@tarr;
    $pkg->{az} = \@az;
    $pkg->{el} = \@el;
}

#sub EL($){
sub EL{
    my $pkg = shift;
    return @{$pkg->{el}};
}

#sub AZ($){
sub AZ{
    my $pkg = shift;
    return @{$pkg->{az}};
}


#sub Tarr($){
sub Tarr{
    my $pkg = shift;
    return @{$pkg->{tarr}};
}

#sub Tsys($$){
sub Tsys{
    my ($pkg,$band) = @_;
    my $ret;
    if($band == 1){# xband
	return @{$pkg->{tsysx}};
    }elsif($band == 2){# sband
	return @{$pkg->{tsyss}};
    }else{
	print "E \@Tsys(): Unexpected band index($band)\n";
	return undef;
    }
}

sub Read_fivept($$){
    my ($pkg, $infile) = @_;

    unless( open( IN, $infile ) ){
	print "E Failed to open infile($infile)\n";
	return -1;
    }
    my $line = <IN>;
    my $mode=0;
    my $npt;
    my (@tmarr, @azfit_err, @elfit_err, @azfit_ofst, @elfit_ofst,
	@azarr, @elarr, @dazarr, @delarr, @srcname, @srcnames,
	@azvarr, @elvarr, @azearr, @elearr,
	@azparam, @elparam);
    my $icnt=0;
    while(! eof(IN)){
	if($line=~/\#fivpt\#fivept/ ){  
	    $mode =1; 
	    my @terms = split(/\s+/,$line);
	    $npt = abs($terms[3]); # N Fivept
	    $pkg->{fivept_npt} = $npt;
	    my (@tdazarr, @tdelarr, @tazvarr, @telvarr, 
		@tazearr, @telearr, @tazparam, @telparam);
	    $dazarr[$icnt] = \@tdazarr;  $delarr[$icnt] = \@tdelarr;
	    $azvarr[$icnt] = \@tazvarr;  $elvarr[$icnt] = \@telvarr;
	    $azearr[$icnt] = \@tazearr;  $elearr[$icnt] = \@telearr;
	    $azparam[$icnt] = \@tazparam;  $elparam[$icnt] = \@telparam;
	    next; 
	}elsif($mode ==1){ 
	    if($line =~/\#fivpt\#lon\s/ ){
		my @terms1 = split(/\s+/,$line);
		my $ix = $terms1[1] -1; # 1-7 -> 0-6
		$dazarr[$icnt][$ix] = $terms1[3];
		$azvarr[$icnt][$ix] = $terms1[4];
		$azearr[$icnt][$ix] = $terms1[5] if(defined($terms1[5]));
	    }elsif($line =~/\#fivpt\#lonfit/ ){
		#lonfit  -0.05500  1.0966 585.965 14.6208  0.0701   5 
		#        offset    HPFW  Peak    base   time_slope
		my @terms1 = split(/\s+/,$line);
		shift(@terms1);
		$azparam[$icnt] = \@terms1;
	    }elsif($line =~/\#fivpt\#lonerr/ ){
		my @terms1 = split(/\s+/,$line);
		$azfit_err[$icnt] = $terms1[1];
	    }elsif($line =~/\#fivpt\#lat / ){
		my @terms1 = split(/\s+/,$line);
		my $ix = $terms1[1] -1; # 1-7 -> 0-6
		$delarr[$icnt][$ix] = $terms1[3];
		$elvarr[$icnt][$ix] = $terms1[4];
		$elearr[$icnt][$ix] = $terms1[5] if(defined($terms1[5]));
	    }elsif($line =~/\#fivpt\#latfit/ ){
		#latfit  -0.01106  0.9780 592.781 20.3491 -0.2931   5
		#        offset    sigma  Peak    base   time_slope
		my @terms1 = split(/\s+/,$line);
		shift(@terms1);
		$elparam[$icnt] = \@terms1;
	    }elsif($line =~/\#fivpt\#laterr/ ){
		my @terms1 = split(/\s+/,$line);
		$elfit_err[$icnt] = $terms1[1];
	    }elsif($line =~/\#fivpt\#offset/ ){
		my @terms1 = split(/\s+/,$line);
		if($terms1[5]*$terms1[6] ==0){ 
		    delete($dazarr[$icnt]);
#		    $mode=0; 
		    next;
		}
		$terms1[0] =~ /(\d{4}.\d{3}).([\d:.]+)\#/;
		my @ydhms = ($1,$2);
		$tmarr[$icnt] = new Time_lib;
		$tmarr[$icnt]->Set_time_strydhms(@ydhms);
		$azarr[$icnt] = $terms1[1];
		$elarr[$icnt] = $terms1[2];
		$azfit_ofst[$icnt] = $terms1[3];
		$elfit_ofst[$icnt] = $terms1[4];
	    }elsif($line =~/\#fivpt\#xoffset/ ){
		my @terms1 = split(/\s+/,$line);
		if( $#terms1 > 9 ){  $srcname[$icnt] = $terms1[10];}
		else{ $srcname[$icnt] = $terms1[9]; }
		push(@srcnames,($terms1[10])) unless(grep(/$terms1[10]/, @srcnames) );
		if($terms1[7]*$terms1[8] ==0){ 
		    delete($dazarr[$icnt]);
		    $mode=0; 
		    next;
		}
		$icnt++;
		$mode=0;
	    }
	}
    }continue{
	$line= <IN>;
    }

    my %fivept=(
	'time_arr',\@tmarr,
	'Az_arr', \@azarr,
	'El_arr', \@elarr,
	'Azfit_err', \@azfit_err,
	'Elfit_err', \@elfit_err,
	'Azfit_ofst', \@azfit_ofst,
	'Elfit_ofst', \@elfit_ofst,
	'Azfit_param', \@azparam,
	'Elfit_param', \@elparam,

	'dAz_arr', \@dazarr,
	'dEl_arr', \@delarr,
	'dAz_Varr', \@azvarr,
	'dAz_Earr', \@azearr,
	'dEl_Varr', \@elvarr,
	'dEl_Earr', \@elearr,
	'Srcname', \@srcname,
	'SrcArr',  \@srcnames,
	);
    
    $pkg->{fivept} = \%fivept;
    $pkg->{fivept_ndat} = $icnt;
    close(IN);
    return $icnt;

}

## tfmt = 0 or undef: Time_lib data
## tfmt = 1: yyyy/ddd-hh:mm:ss
## 
sub Fivept_Tdat() {
    my $pkg1= shift;
    my $ix= shift;
    my $tfmt= shift;
    my $fivept =  $pkg1->{fivept};
    my $tarr =  $fivept->{time_arr};
    my $ret;
    if( $tfmt == 1){
	unless(defined($$tarr[$ix])){
	    printf STDERR "E FSlog::Fivept_Tdat(): tarr[%d] undefined.\n", $ix;
	    return "InValid";
	}
	$ret = $$tarr[$ix]->Str_fmt1($tfmt);
    }elsif(defined($tmft) || $tmft ==0){
	$ret = $$tarr[$ix];
    }
    return $ret;
}

sub Fivept_Azarr(){
    my $pkg= shift;
    my $fivept =  $pkg->{fivept};
    my $arr = $fivept->{Az_arr};
    return $arr; 
}
sub Fivept_Elarr(){
    my $pkg= shift;
    my $fivept =  $pkg->{fivept};
    my $arr =  $fivept->{El_arr};
    return $arr; 
}
#sub Fivept_fitparam($$){
sub Fivept_fitparam(){
    my ($pkg,$ix)= @_;
    my $Azfit =  ${$pkg->{fivept}}->{Azfit_ofst};
    my $Elfit =  ${$pkg->{fivept}}->{Elfit_ofst};
    my $Azfit_err =  ${$pkg->{fivept}}->{Azfit_err};
    my $Elfit_err =  ${$pkg->{fivept}}->{Elfit_err};
    my @ret = (	$$Azfit[$ix],$$Elfit[$ix],
		$$Azfit_err[$ix],$$Elfit_err[$ix]);
    return @ret;
}
sub Fivept_AzValarr($){
    my $pkg= shift;
    my $fivept =  $pkg->{fivept};
    my $arr =  $fivept->{dAz_Varr};
    return $arr; }
sub Fivept_ElValarr($){
    my $pkg= shift;
    my $fivept =  $pkg->{fivept};
    my $arr =  $fivept->{dEl_Varr};
    return $arr; }

sub Fivept_dAzarr(){
    my $pkg= shift;
    my $fivept =  $pkg->{fivept};
    my $arr =  $fivept->{dAz_arr};
    return $arr; 
}
sub Fivept_dElarr(){
    my $pkg= shift;
    my $fivept =  $pkg->{fivept};
    my $arr =  $fivept->{dEl_arr};
    return $arr; 
}

##  Fivept Ngood data
sub Fivept_Ndat(){
    my $pkg= shift;
    return $pkg->{fivept_ndat};   
}

##  Fivept points
sub Fivept_npt(){
    my $pkg= shift;
    return $pkg->{fivept_npt};   
}
sub Fivept_AzFitparam( $$ ){
    my $pkg= shift;
    my $ix= shift;
    my $fivept = $pkg->{fivept};
    my $azparam = $fivept->{Azfit_param};   
    return @{$$azparam[$ix]};
}
sub Fivept_ElFitparam( $$ ){
    my $pkg= shift;
    my $ix= shift;
    my $fivept = $pkg->{fivept};
    my $param = $fivept->{Elfit_param};   
    return @{$$param[$ix]};
}


sub Fivept_Tarr {
    my $pkg= shift;
    my $fivept = $pkg->{fivept};
    my $tarr =  $fivept->{time_arr};
    return $tarr;
}

# Source name for each scan
sub Fivept_Srcarr {
    my $pkg= shift;
    my $fivept = $pkg->{fivept};
    my $arr =  $fivept->{Srcname};
    return $arr;
}

# Source names 
sub Fivept_Srcnames {
    my $pkg= shift;
    my $fivept = $pkg->{fivept};
    my $arr =  $fivept->{SrcArr};
    return $arr;
}

1;
