# @Version: $Id: GeoCord.pm 89 2016-04-19 12:44:37Z sekido $
#
package GeoCord;
#use Exporter;
use Math::Matrix;
use vectopr;

BEGIN{
    $EarthMajorRadius = 6378137.;  # [m]
    $EarthFlattening = 1./298.257222101;  ##
    $PI = 4.*atan2(1,1);
    $deg2rad = $PI / 180.;
    $eccentricity = sqrt(2*$EarthFlattening 
			 - $EarthFlattening*$EarthFlattening);
}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

 ## 卯酉線 ぼうゆう線 曲率半径
## Curvature Raidus
sub PrimeVerticalRaidus($$){
    my $pkg = shift;
    my $lat = shift;
    my $rlat = $lat * $deg2rad;
    my $ps = sin($rlat);
    my $tmp =  $eccentricity * $eccentricity *$ps*$ps;
    my $N = $EarthMajorRadius/sqrt(1 - $tmp );
    return $N;
}

 ## 子午線 しご線 曲率半径
## Curvature Raidus
sub MeridianRaidus($$){
    my $pkg = shift;
    my $lat = shift;
    my $rlat = $lat * $deg2rad;
    my $ps = sin($rlat);
    my $e2 = $eccentricity * $eccentricity;
    my $tmp =  $e2 *$ps*$ps;
    my $M = $EarthMajorRadius* (1. - $e2)/(1 - $tmp );
    return $M;
}

## Lon, Lat [deg.], Height [m] 
sub Llh2XYZ($$$$){
    my ($pkg, $lon, $lat, $h, $hgeoid) = @_;
    my $ph = $lat * $deg2rad;
    my $lamd = $lon * $deg2rad;
    my $cp = cos($ph);
    my $sp = sin($ph);
    my $cl = cos($lamd);
    my $sl = sin($lamd);
    
    my $N = $pkg->PrimeVerticalRaidus($lat); 
    my $h1 =$h + $hgeoid + $N;
    my $e2 = $eccentricity * $eccentricity;
    my @xyz = (
	$h1 * $cp * $cl,
	$h1 * $cp * $sl,
	($N * (1. - $e2) + $h + $hgeoid) * $sp
	);
    return @xyz
}

## Input XYZ [m]
## Return Lon/Lat [deg], height [m]
sub XYZ2llh($$){
    my ($pkg,$xyz) = @_;
    my $P = sqrt($$xyz[0][0] * $$xyz[0][0] + $$xyz[1][0] * $$xyz[1][0]);
    my $e2 = $eccentricity * $eccentricity;
    my $lat0 = atan2($$xyz[2][0],$P) / $deg2rad;
    my $lat = $lat0;
    my $N = $pkg->PrimeVerticalRaidus($lat); 
    my $imprv = 1;
    while($imprv > 1.e-9){
	my $cp = cos($lat*$deg2rad);
	my $tmp = ($P-$e2*$N*$cp);
	my $lat1 = atan2($$xyz[2][0], $tmp)/$deg2rad;
	my $N1 = $pkg->PrimeVerticalRaidus($lat1); 
	my $dl = $lat1 - $lat;
	$imprv = abs($dl/$lat);
#	printf "Lat (%s) Improve = %.9f\n",deg2str($lat1), $imprv;
	$lat = $lat1; $N = $N1;
    }
    my $lon = atan2($$xyz[1][0], $$xyz[0][0])/$deg2rad;
    my $h = $P/cos($lat*$deg2rad) - $N;

    return ($lon,$lat,$h);
}


## Converting XYZ to ENH by rotation 
## XYZ in [m]
## dxyz and XYZ are row vectors.
sub XYZ2ENH($$$){
    my ($pkg, $xyz, $dxyz) = @_;
#    my @llh = $pkg->XYZ2llh($xyz);
    my $r = sqrt($$xyz[0][0]*$$xyz[0][0] + $$xyz[1][0]*$$xyz[1][0]);
    my @llh = ( atan2($$xyz[1][0],$$xyz[0][0])/$deg2rad, 
		 atan2($$xyz[2][0],$r)/$deg2rad  );

    my $mz = vectopr::Rz($llh[0]);
    my $my = vectopr::Ry(90 - $llh[1]);
    my $mz2 =vectopr::Rz(90);
    my $enh = $mz2 * $my * $mz * $dxyz;
    return $enh;
}
## Converting ENH to XYZ by rotation 
## XYZ in [m]
## dxyz and XYZ are row vectors.
sub ENH2XYZ($$$){
    my ($pkg, $xyz, $enh) = @_;
#    my @llh = $pkg->XYZ2llh($xyz);
    my $r = sqrt($$xyz[0][0]*$$xyz[0][0] + $$xyz[1][0]*$$xyz[1][0]);
    my @llh = ( atan2($$xyz[1][0],$$xyz[0][0])/$deg2rad, 
		 atan2($$xyz[2][0],$r)/$deg2rad  );

    my $mz = vectopr::Rz( -$llh[0]);
    my $my = vectopr::Ry($llh[1] - 90);
    my $mz2 =vectopr::Rz(-90);
    my $dxyz = $mz * $my * $mz2 * $enh;
    return $dxyz;
}


sub deg2str($){
    my $lon = shift;
    my $s = $lon > 0 ? 1 : -1;

    my $dlon = $s * $lon;
    my $d = int($dlon);
    my $rsd = ($dlon - $d)*60;
    my $m = int($rsd);
    my $sec = ($rsd - $m)*60;
    sprintf("%+3dd %02dm %08.5f",$d*$s, $m,$sec);
}

sub hour2str($){
    my $lat = shift;
    my $s = $lat > 0 ? 1 : -1;

    my $dlat = $s * $lat;
    my $d = int($dlat);
    my $rsd = ($dlat - $d)*24;
    my $m = int($rsd);
    my $sec = ($rsd - $m)*60;
    sprintf("%+3dh %02dm %08.5f",$d*$s, $m,$sec);
}


1;


