# @Version: $Id: Gico_dat.pm 121 2017-04-22 07:56:24Z vlbi $
#
package Gico_dat;
use Exporter;
use Carp;

use Time::Local;
use Time_lib;
@ISA = qw(Exporter);

BEGIN{
}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

## Reading fist 1 set of data from GICO file
sub Read1($$){
    my $pkg = shift;
    my $infile = shift;
    
    unless(open(IN, $infile)){
	print "Error Gico_dat: faile to open gico file ($infile)\n";
	return -1;
    }
    my @list = <IN>;
    close(IN);
    
    my $sum_flag=0;
    foreach $l (@list){
	my $chk_mask=0;
	my @terms = split(/\s+/,$l);
	my $key1 = substr(shift(@terms),1);
	if($key1 =~/STATION/ ||  $key1 =~/SOURCE/   ){
	    $key1 =~ s/-/_/ if($key1=~/STATION/);
	    $pkg->{$key1} = \@terms;
	    $chk_mask = 0x01 if($key1 eq "STATION-X");
	    $chk_mask = 0x02 if($key1 eq "STATION-Y");
	    $chk_mask = 0x04 if($key1 eq "SOURCE");
	}elsif( ($key1 =~m!STIME! ) || ($key1 =~m!ETIME!) ||  ($key1 =~m!PRT!) ){
	    my $t1 = new Time_lib;
	    $t1->Set_time_strydhms(@terms);
	    $pkg->{$key1} = $t1;
	    $chk_mask = 0x08 if($key1 eq "STIME");
	    $chk_mask = 0x10 if($key1 eq "ETIME");
	    $chk_mask = 0x20 if($key1 eq "PRT");
	}else{
	    $pkg->{$key1} = $terms[0];
	}
	if($chk_mask & $sum_flag){ last;}
	else{ $sum_flag |= $chk_mask;}
	## return if the second entry appears.
    }
    return 0;
}

sub Source($){   $pkg = shift;   @{$pkg->{SOURCE}};}
sub Exp($){   $pkg = shift;   $pkg->{EXPER};}
sub Scan($){   $pkg = shift;   $pkg->{SCAN};}
sub Stime($){   $pkg = shift;   $pkg->{STIME};}
sub Etime($){   $pkg = shift;   $pkg->{ETIME};}
sub PRT($){   $pkg = shift;   $pkg->{PRT};}
sub Xst($){ $pkg = shift; my $arr = $pkg->{STATION_X}; return @{$arr};}
sub Xstname($){ $pkg = shift; my $arr = $pkg->{STATION_X}; return $$arr[1];}
sub XstID($){ $pkg = shift; my $arr = $pkg->{STATION_X}; return $$arr[0];}
sub XstXYZ($){ $pkg = shift;   my $arr = $pkg->{STATION_X};
	       return ($$arr[2],$$arr[3],$$arr[4]);  }

sub Yst($){ $pkg = shift; my $arr = $pkg->{STATION_Y}; return @{$arr};}
sub Ystname($){ $pkg = shift; my $arr = $pkg->{STATION_Y}; return $$arr[1];}
sub YstID($){ $pkg = shift;  my $arr = $pkg->{STATION_Y};  return $$arr[0];}
sub YstXYZ($){  $pkg = shift; my $arr = $pkg->{STATION_Y};
		return ($$arr[2],$$arr[3],$$arr[4]); }

sub COffset($){ $pkg = shift;  return $pkg->{CLKOFFSET};}
sub CRate($){ $pkg = shift;  return $pkg->{CLKRATE};}
sub Apr0($){ $pkg = shift;  return $pkg->{APR0};} # sec
sub Apr1($){ $pkg = shift;  return $pkg->{APR1};} #s/s
sub Apr2($){ $pkg = shift;  return $pkg->{APR2};} # s/ss
sub Apr3($){ $pkg = shift;  return $pkg->{APR3};} #s/sss
## Argument: Delta t for PRT
sub Apr0t($$){
    my $pkg = shift;   my $dt = shift;
    my $apr0 = $pkg->{APR0};
    my $crr1 = $pkg->{APR1} * $dt;
    my $crr2 = 0.5*$pkg->{APR2} * $dt * $dt;
    my $crr3 = $pkg->{APR3} * $dt * $dt * $dt/6.0;
    my $apr0t = $apr0 + $crr1 +  $crr2 +  $crr3;
    return $apr0t;
}
## dtau/dt Argument: Delta t for PRT
sub Apr1t($$){
    my $pkg = shift;   my $dt = shift;
    my $apr1 = $pkg->{APR1};
    my $crr1 = $pkg->{APR2} * $dt;
    my $crr2 = 0.5*$pkg->{APR3} * $dt * $dt;
    my $apr1t = $apr1 + $crr1 +  $crr2;
    return $apr1t;
}
## d2tau/dt^2 Argument: Delta t for PRT
sub Apr2t($$){
    my $pkg = shift;   my $dt = shift;
    my $apr2 = $pkg->{APR2};
    my $crr1 = $pkg->{APR3} * $dt;
    my $apr2t = $apr2 + $crr1;
    return $apr2t;
}
sub Freq($){ $pkg = shift;  return $pkg->{FREQUENCY};} # Hz
sub Srate($){ $pkg = shift;  return $pkg->{SAMPLE};} # Hz
sub PP($){ $pkg = shift;  return $pkg->{PP};} # sec
sub Tinteg($){ $pkg = shift;  return $pkg->{TINTEG};} # sec
sub dTau($){ $pkg = shift;  return $pkg->{DELAY};} # sec
sub dTaut($$){ $pkg = shift;  $dt = shift; 
	       my $crr1 = $pkg->{RATE} * $dt;
	       return $pkg->{DELAY} + $crr1;  } # sec
# whole delay:Apri + residual 
sub ObsTau($){ $pkg = shift;  my $del =  $pkg->{DELAY} +  $pkg->{APR0}; 
	       return $del;} # sec
sub dRate($){ $pkg = shift;  return $pkg->{RATE};} # sec/sec
# whole Rate :Apri + residual 
sub ObsRate($){ $pkg = shift;  my $del =  $pkg->{RATE} +  $pkg->{APR1}; 
	       return $del;} # sec
sub Amp($){ $pkg = shift;  return $pkg->{AMPLITUDE};} # 
sub Phase($){ $pkg = shift;  return $pkg->{PHASE};} # deg.
sub SNR($){ $pkg = shift;  return $pkg->{SNR};} # 
sub Value($$){
    my $pkg = shift;
    my $key = shift;
    return $pkg->{$key};
}

sub Set_StnX($$){   $pkg = shift;  $pkg->{STATION_X} = shift;}
sub Set_StnY($$){   $pkg = shift;  $pkg->{STATION_Y} = shift;}
sub Set_APR($$$$$){   
    $pkg = shift;  
    $pkg->{APR0} = shift;
    $pkg->{APR1} = shift;
    $pkg->{APR2} = shift;
    $pkg->{APR3} = shift;
}
sub Set_Tau($$$){   
    $pkg = shift;  
    $pkg->{DELAY} = shift;
    $pkg->{RATE} = shift;
}
sub Set_AmpPhase($$$){   
    $pkg = shift;  
    $pkg->{AMPLITUDE} = shift;
    $pkg->{PHASE} = shift;
}
sub Set_SNR($$){   
    $pkg = shift;  
    $pkg->{SNR} = shift;
}
sub Set_Clock($$$){   
    $pkg = shift;  
    $pkg->{CLKOFFSET} = shift;
    $pkg->{CLKRATE} = shift;
}


sub Write($$){
    my $pkg = shift;
    my $ofile = shift;
    unless( open(OUT, ">$ofile")){
	print STDERR "Error Gico_dat: Failed to open outfile($ofile)\n";
	return -1;
    }
    printf OUT "\$EXPER          %s\n",$pkg->{EXPER};
    printf OUT "\$SCAN           %04d\n",$pkg->{SCAN};
    printf OUT "\$STATION-X      %s %s %s %s %s\n",@{$pkg->{STATION_X}};
    printf OUT "\$STATION-Y      %s %s %s %s %s\n",@{$pkg->{STATION_Y}};
    printf OUT "\$SOURCE         %-8s %s %s %s %s %s %s %s\n",@{$pkg->{SOURCE}};
    printf OUT "\$STIME          %s\n",$pkg->{STIME}->Str_fmt1(0);
    printf OUT "\$ETIME          %s\n",$pkg->{ETIME}->Str_fmt1(0);
    printf OUT "\$PRT            %s\n",$pkg->{PRT}->Str_fmt1(0);
    printf OUT "\$UT1UTC         %+21.14e\n",$pkg->{UT1UTC};
    printf OUT "\$CLKOFFSET      %+21.14e\n",$pkg->{CLKOFFSET};
    printf OUT "\$CLKRATE        %+21.14e\n",$pkg->{CLKRATE};
    printf OUT "\$APR0           %+21.14e\n",$pkg->{APR0};
    printf OUT "\$APR1           %+21.14e\n",$pkg->{APR1};
    printf OUT "\$APR2           %+21.14e\n",$pkg->{APR2};
    printf OUT "\$APR3           %+21.14e\n",$pkg->{APR3};
    printf OUT "\$FREQUENCY      %+.6f\n",$pkg->{FREQUENCY};
    printf OUT "\$SAMPLE         %.3f\n",$pkg->{SAMPLE};
    printf OUT "\$PP             %+.6f\n",$pkg->{PP};
    printf OUT "\$TINTEG         %+.3f\n",$pkg->{TINTEG};
    printf OUT "\$DELAY          %+21.14e\n",$pkg->{DELAY};
    printf OUT "\$RATE           %+21.14e\n",$pkg->{RATE};
    printf OUT "\$AMPLITUDE      %+15.8e\n",$pkg->{AMPLITUDE};
    printf OUT "\$PHASE          %+.3f\n",$pkg->{PHASE};
    printf OUT "\$SNR            %+.3f\n",$pkg->{SNR};
    close(OUT);
}



1;
