## Metrorogical data reading module.
## This data is taken from JMAP web site at 
## http://www.data.jma.go.jp/obd/stats/etrn/
## data format is expected to be 
#		ʿ		ִ	
## Time  P(local) P(sea) Rain[mm] Temp[deg.] Humidity[%] Wind [m/s] Directions
# 00:10	1018.9	1022.3	0.0	9.3	88	1.2		1.9		

package JMAMet_dat;
use Exporter;
use Carp;

use Time_lib;

@ISA = qw(Exporter);

BEGIN{
    %wdc =(
	'' ,'N',
	'' ,'NNE',
	'' ,'NE',
	'' ,'ENE',
	'' ,'E',
	'' ,'ESE',
	'' ,'SE',
	'' ,'SSE',
	'' ,'S',
	'' ,'SSW',
	'' ,'SW',
	'' ,'WSW',
	'' ,'W',
	'' ,'WNW',
	'' ,'NW',
	'' ,'NNW'
	);
    
}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

sub Read($$$){
    my $pkg = shift;
    my $infile = shift;
    my $t0 = shift;

    
    unless(open(IN,$infile)){
	print STDERR "E Failed to Met_dat::Read() \n";
	return -1;
    }
    my (@t,@p,@h, @wsm, @wd,@wsx, @tarr);
    my $yy = $t0->Year();
    my $doy = $t0->Doy();
    
    my $icnt = 0;
    my $tm_prev;
    while(<IN>){
	next if(/^\#/);
	my @terms = split(/[\(\)\s\t]+/);
	next if(@terms < 7);
	my @hm = split(/:/,$terms[0]);
	my @ydhms = ( $yy,$doy,@hm, 0 );
	$tarr[$icnt] = new Time_lib;
	$tarr[$icnt]->Set_time_ydhms(@ydhms);
	$tarr[$icnt] -= 9 * 3600;  # Because Time tag is JST(+9) 
	$t[$icnt] = $terms[4];
	$p[$icnt] = $terms[1];
	$h[$icnt] = $terms[5];
	$wsm[$icnt] = $terms[6];
	$wd[$icnt] = $wdc{$terms[7]};
	$wsx[$icnt] = $terms[8];
	$icnt++;
    }
    $pkg->{temp}  = \@t;
    $pkg->{pres}  = \@p;
    $pkg->{humid} = \@h;
    $pkg->{WndSpeedMedian} = \@wsm;
    $pkg->{WndDirection} = \@wd;
    $pkg->{WndSpeedMax} = \@wsx;
    $pkg->{time}  = \@tarr;
    close(IN);
    return ($#t +1);
}

sub WSpeedMedian($){
    my $pkg = shift;
    return @{$pkg->{WndSpeedMedian}};
}
sub WSpeedMax($){
    my $pkg = shift;
    return @{$pkg->{WndSpeedMax}};
}
sub WDirection($){
    my $pkg = shift;
    return @{$pkg->{WndDirection}};
}

sub Temp($){
    my $pkg = shift;
    return @{$pkg->{temp}};
}

sub Pres($){
    my $pkg = shift;
    return @{$pkg->{pres}};
}

sub Humid($){
    my $pkg = shift;
    return @{$pkg->{humid}};
}

sub Tarr($){
    my $pkg = shift;
    return @{$pkg->{time}};
}

1;
