# @Version: $Id: Phcon.pm 48 2014-11-30 03:51:32Z sekido $
#
package Phcon;
#
# Expected format is 
# # yd hms mjd tau rambg nambg amp derr snr
# 2005/322 21:36:01.000 53692.9000115699964 +2.84423781013e-08 +0.0000000e+00       0  1352-104 8337418571.4000  1.9068e-03 -8.78e+01 140.30
# 
use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(Read Rrite Tarr Ndat);

# use PGPLOT;

BEGIN {
    ;
}


sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

sub Ndat(){
    $this->{ndat};
}

sub Tarr($$){
    my ($this, $idat) = @_;
    
    ${$this->{tarr}}[$idat];
}

sub Write($$){
    my ($this, $outfile) = @_;
    my $ndat = $this->{ndat};
    my $tarr = $this->{tarr};
    my $mjd = $this->{mjd};
    my $frq = $this->{frq};
    my $amb = $this->{amb};
    my $iamb = $this->{iamb};
    my $src = $this->{src};
    my $dtau = $this->{dtau};
    my $amp = $this->{amp};
    my $err = $this->{err};
    my $snr = $this->{snr};

    open(OUT, ">$outfile") || die "Error: Cannot open file $outfile \n";
    for(my $i=0; $i < $ndat; $i++){
	$$tarr[$i]->Print_fmt1(OUT);
	printf OUT " %19.13f %+18.11e %13.7e %6d %8s %15.4f %10.4e %8.2e %6.2f\n",
	$$tarr[$i]->Mjd(), $$dtau[$i], $$amb[$i], $$iamb[$i], $$src[$i], 
	$$frq[$i], $$amp[$i], $$err[$i], $$snr[$i];
# mjd[$i], $$dtau[$i], $$amb[$i], $$iamb[$i], $$src[$i], 

	if($i<10){
	    $$tarr[$i]->Print_fmt1();
	    printf STDOUT " %13f %+18.11e %13.7e %6d %8s %15.4f %10.4e %8.2e %6.2f\n",
	    $$tarr[$i]->Mjd(), $$dtau[$i], $$amb[$i], $$iamb[$i], $$src[$i], 
	    $$frq[$i], $$amp[$i], $$err[$i], $$snr[$i];
	}

    }
    close(OUT);
    
}

## Write L rate & rerr
sub Write_ext1($$){
    my ($this, $outfile) = @_;
    my $ndat = $this->{ndat};
    my $tarr = $this->{tarr};
    my $mjd = $this->{mjd};
    my $frq = $this->{frq};
    my $amb = $this->{amb};
    my $iamb = $this->{iamb};
    my $src = $this->{src};
    my $dtau = $this->{dtau};
    my $amp = $this->{amp};
    my $err = $this->{err};
    my $snr = $this->{snr};
    my $rat = $this->{rat};
    my $rerr = $this->{rerr};

    open(OUT, ">$outfile") || die "Error: Cannot open file $outfile \n";
    for(my $i=0; $i < $ndat; $i++){
	$$tarr[$i]->Print_fmt1(OUT);
	printf OUT " %19.13f %+18.11e %13.7e %6d %8s %15.4f %10.4e %8.2e %6.2f %13.7e %8.2e\n",
	$$mjd[$i], $$dtau[$i], $$amb[$i], $$iamb[$i], $$src[$i], 
	$$frq[$i], $$amp[$i], $$err[$i], $$snr[$i], $$rat[$i], $$rerr[$i];

	if($i<10){
	    $$tarr[$i]->Print_fmt1();
	    printf STDOUT " %13f %+18.11e %13.7e %6d %8s %15.4f %10.4e %8.2e %6.2f %13.7e %8.2e\n",
	    $$mjd[$i], $$dtau[$i], $$amb[$i], $$iamb[$i], $$src[$i], 
	    $$frq[$i], $$amp[$i], $$err[$i], $$snr[$i], $$rat[$i], $$rerr[$i];
	}

    }
    close(OUT);
    
}

sub Read($$){
    my ($this, $infile) = @_;

    open(IN, $infile) || die "Error: Cannot open file $infile \n";
    
    my (@mjd, @yd, @hms, @dtau, @frq, @src, @amp, @amb, @iamb, @tarr);
    my $cnt=0;
    while(<IN>){
	next if(/^\#/);
	/\s*/;
	my $line = $'; # '
	my @terms = split(/\s+/,$line );
	if( @terms < 10){
	    print "Error: phcon file format is wrong\n";
	    print "We expect: yd hms mjd tau rambg, nambg, src frq amp err snr\n";
	    print "line= $line \n";
	    exit(1);
	}

	$yd[$cnt] = $terms[0];
	$hms[$cnt] = $terms[1];
	$mjd[$cnt] = $terms[2];
	$dtau[$cnt] = $terms[3];
	$amb[$cnt] = $terms[4];
	$iamb[$cnt] = $terms[5];
	$src[$cnt] = $terms[6];	
	$frq[$cnt] = $terms[7];
	$amp[$cnt] = $terms[8];
	$err[$cnt] = $terms[9];
	$snr[$cnt] = $terms[10] if(defined($terms[10]));
	$tarr[$cnt] = new Time_lib;
	my @tmp1 = split(m!/!,$yd[$cnt]);
	my @tmp2 = split(m!:!,$hms[$cnt]);
	$tarr[$cnt]->Set_time_ydhms(@tmp1,@tmp2); 
	$cnt++;
    }
    close(IN);

    $this->{mjd} = \@mjd;
    $this->{tarr} = \@tarr;
    $this->{yd} = \@yd;
    $this->{hms} = \@hms;
    $this->{dtau} = \@dtau;
    $this->{frq} = \@frq;
    $this->{src} = \@src;
    $this->{amp} = \@amp;
    $this->{snr} = \@snr;
    $this->{err} = \@err;    
    $this->{amb} = \@amb;
    $this->{iamb} = \@iamb;

    $this->{ndat} = $cnt;

    return($cnt);
}

sub Set_mjdarr($$){
    my ($this,$arr) = @_;
    $this->{mjd} = $arr;
}
sub Set_tarr($$){
    my ($this,$arr) = @_;
    $this->{tarr} = $arr;
}
sub Set_dtau($$){
    my ($this,$arr) = @_;
    $this->{dtau} = $arr;
}
sub Set_snr($$){
    my ($this,$arr) = @_;
    $this->{snr} = $arr;
}
sub Set_err($$){
    my ($this,$arr) = @_;
    $this->{err} = $arr;
}
sub Set_frqarr($$){
    my ($this,$arr) = @_;
    $this->{frq} = $arr;
}
sub Set_src($$){
    my ($this,$arr) = @_;
    $this->{src} = $arr;
}
sub Set_amp($$){
    my ($this,$arr) = @_;
    $this->{amp} = $arr;
}
sub Set_iamb($$){
    my ($this,$arr) = @_;
    $this->{iamb} = $arr;
}
sub Set_amb($$){
    my ($this,$arr) = @_;
    $this->{amb} = $arr;
}
sub Set_ndat($){
    my ($this) = @_;
    my $tarr = $this->{tarr};
    my $ndat = @$tarr;
    $this->{ndat} = $ndat;
}
sub Set_rat($$){
    my ($this,$arr) = @_;
    $this->{rat} = $arr;
}
sub Set_rerr($$){
    my ($this,$arr) = @_;
    $this->{rerr} = $arr;
}
