# @Version: $Id: RA1631.pm 48 2014-11-30 03:51:32Z sekido $
#
package RA1631;
use Exporter;
@ISA = qw(Exporter);

use Mathfnc;

@EXPORT = qw(Set_param Model_a Model_b);

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

sub Set_param($$$){
    my ($this,$frq, $diam) = @_;
    my $VLIGHT = 299792458.;
    my $pi = atan2(1.,1.) *4.0;
    my $r2d = 180./$pi;

    my $lambda = $this->{lambda} = $VLIGHT/$frq;
    $this->{frq} = $frq;   ## (Hz)
    $this->{diameter} = $diam;
    my $dl_ratio = $this->{dl_ratio} = $diam / $lambda ; # radian 
    my $ld_ratio = 1./$dl_ratio;
    
    my $gmax = $this->{Gmax_a} = 
	20.* ( log($dl_ratio) + log($pi))/log(10.); ## dBi 
    my $g1 = $this->{G1} = -1 + 15.*log($dl_ratio)/log(10); ## dBi
    my $w = $pi * $dl_ratio;
    $this->{Gmax_b} = $w*$w;

    $this->{phai_m} = 20.*$ld_ratio * sqrt($gmax - $g1) ; # deg.
    $this->{phai_r} = 15.85* ($ld_ratio **0.6) ; ## deg. 
    $this->{phai_0} =69.88/$dl_ratio; # deg.
}

## Precise model for phai < 1 deg.
sub Model_b($$){
    my ($this,$phai) = @_;
    my $aphai = abs($phai);
    my $pi = atan2(1.,1.) *4.;
    my $dl_ratio = $this->{dl_ratio};
    my $x = $pi /360. * $dl_ratio * $aphai; ## rad;
    my $gmax = $this->{Gmax_b};
    my $phai_0 = $this->{phai_0}; # deg.

    my ($b, $w);
    my $ret;
    if($phai < $phai_0){
	$b=$gmax;
	$w = $x ==0 ? 1 : &Bssl1(2*$pi*$x)/($pi*$x);
    }else{
	$b = (10**3.2) * ($pi* $pi *$dl_ratio*0.5/180.)**2;
	$w = cos(2.*$pi*$x - 3.*$pi*0.25 + 0.0953)/($pi*$x);
    }
    return( $b * $w**2);
}

## rough model 
sub Model_a($$){
    my ($this,$phai) = @_;
    my $aphai = abs($phai);
    my $pi = atan2(1.,1.) *4.;
    my $aphai_rad = $aphai * $pi / 180.;

    if( $aphai < $this->{phai_m}){
	my $dl_ratio = $this->{dl_ratio};
	my $w = 2.5e-3 * ($dl_ratio * $aphai)**2;
	return($this->{Gmax_a} - $w);
    }elsif( $aphai < $this->{phai_r}){
	return($this->{G1});
    }elsif(  $aphai < 10){
	return(29-25*log($aphai)/log(10.));
    }elsif(  $aphai < 34.1){
	return(34 - 30*log($aphai)/log(10.));
    }elsif(  $aphai < 80){
	return(-12);
    }elsif(  $aphai < 120){
	return(-7);
    }elsif(  $aphai < 180){
	return(-12);
    }

}
