# Computing Radio Flux data of Tausus-A, Virgo-A, Cygnus-A, Cassiopea-A, Sagittarius-A
# Rev: $Rev: 54 $ $Id: RadioFlux.pm 54 2014-12-08 09:13:36Z sekido $
# based on 
# http://www.eme2008.org/papers/Sistem%20components/Radio_Sources_Eng.pdf
# http://www.monstein.de/astronomypublications/TaurusA/ORIONTAU.htm
# http://www.monstein.de/astronomypublications/VirgoA/virgo.pdf
package RadioFlux;

BEGIN{
    $Jy = 1.e-26;
}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

# Return Flux in Jy
# Input Freq. in [MHz]
sub TauA($$){
    my $pkg = shift;
    my $f = shift;
#    my $sdb = -220.85 - 2.99 * log($f)/log(10.0);
    my $sdb = 3.915 - 0.299 * log($f)/log(10.0);
#    $sdb +=260;
    my $s = 10**($sdb);
    return $s;
}


# Return Flux in Jy
# Input Freq. in [MHz]
sub VirgoA($$){
    my $pkg = shift;
    my $f = shift;
#    my $a = -209.77;
#    my $b = -8.56;
#    my $sdb = $a + $b * log($f)/log(10.0);
    my $sdb = 5.023 -0.856 * log($f)/log(10.0);
#    $sdb +=260;
    my $s = 10**($sdb);
    return $s;
}

# Return Flux in Jy
# Input Freq. in [MHz]
sub CygnusA($$){
    my $pkg = shift;
    my $f = shift;
#    my ($a,$b,$c) = (0,0,0);
    my $x=log($f)/log(10.0);
    my $sdb=0;
    if($f < 2000){
#	$a = -213.05;
#	$b = 0.85;
#	$b = -1.78;
	$sdb = 4.695+0.085*$x -0.178*$x*$x;
    }else{
#	$a = -225.71;
#	$b =  -12.44;
	$sdb = 7.161 -1.244*$x;
    }

#    my $sdb = $a + $b * $x + $c * $x *$x;
#    $sdb +=260;
    my $s = 10**($sdb);
    return $s;
}

# Return Flux in Jy
# Input Freq. in [MHz], and Year yyyy
sub CasA($$$){
    my $pkg = shift;
    my $f = shift;
    my $year = shift;
    my $x = log($f)/log(10);
    my $a = 5.745 -0.77*$x;
    my $b = (0.0187 - 0.003*$x) * ($year - 1980);
    my $s = (1.0- $b) * 10**($a);
    return $s;
}

1;
