# @Version: $Id: Time_lib.pm 81 2016-01-21 10:56:01Z vlbi $
#
package Sinex;
#use Exporter;
#use Carp;
use Time_lib;
use GeoCord;

BEGIN{

}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    return $this;
}

sub Get_Site_Id( $$ ){
    my ($pkg, $infile) = @_;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SITE/ID!){
	    $mode =1; # found SITE/ID
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_Site_Id():  SITE/ID not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@sitecode, @pntcode, @MonuID, @obscode, @sitedesc, @lon, @lat, @height);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SITE/ID!){  $mode = 2; next; }
	$sitecode[$icnt] = substr($line,1,4);
	$pntcode[$icnt] = substr($line,6,2);
	$MonuID[$icnt] = substr($line,9,9);
	$obscode[$icnt] = substr($line,19,1);
	$sitedesc[$icnt] = substr($line,21,22);
	my $l1 = substr($line,44,3);
	my $sg = $l1 > 0 ? 1 : -1;  $l1 *= $sg;
	my $l2 = substr($line,48,2);
	my $l3 = substr($line,51,4);
	$lon[$icnt] = ($l3/60. + $l2)/60. + $l1; 
	$lon[$icnt] *= $sg;

	$l1 = substr($line,56,3);
	$sg = $l1 > 0 ? 1 : -1;  $l1 *= $sg;
	$l2 = substr($line,60,2);
	$l3 = substr($line,63,4);
	$lat[$icnt] = ($l3/60. + $l2)/60. + $l1; 
	$lat[$icnt] *= $sg;
	
	$height[$icnt] = substr($line,68,7);
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    my %siteid= (
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'MonumentID', \@MonuID,
	'ObsCode', \@obscode,
	'SiteDescr', \@sitedesc,
	'Longitude', \@lon,
	'Latitude', \@lat,
	'Height', \@height,
	'Nsite', $icnt);
    $pkg->{siteid} = \%siteid;
    close(IN);
    return $icnt;
}

# Number of Sites
sub Nsite($){
    my $pkg= shift;
    $siteid = $pkg->{siteid};
    return $siteid->{Nsite};
}

sub SiteDescs($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{SiteDescr};
}

# Return Site-Description of the sitecode
sub SiteDesc($$){
    my $pkg = shift;
    my $stcode = shift;
    my $sitedesc = $pkg->{siteid}->{SiteDescr};
    my $sitecode = $pkg->{siteid}->{SiteCode};
    for(my $i=0; $i<@$sitecode; $i++){
	if($$sitecode[$i] =~/$stcode/ ){
	    return $$sitedesc[$i];
	}
    }
    return ""; 
}

sub SiteCodes($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{SiteCode};
}

## Find pattern in the site description
sub SiteCodes_DescPtn($$){
    my $pkg = shift;
    my $ptn = shift;
    $ptn = lc($ptn);

    my $siteid = $pkg->{siteid};
    my $siteDesc = $siteid->{SiteDescr};
    my $siteCode = $siteid->{SiteCode};
    my $found =0;
    my @ids =();
    for(my $i=0; $i< @$siteDesc; $i++){
	if( lc($$siteDesc[$i]) =~/$ptn/ ){
	    $ids[$found] = $$siteCode[$i] ;
#	    push(@ids, ( $$siteCode[$i]));
	    $found++;
	}
    }
    return @ids;
}


# Longitude
sub SiteLon($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{Longitude};
}

# Latitude
sub SiteLat($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{Latitude};
}


sub Get_Sol_Epoch( $$ ){
    my ($pkg, $infile) = @_;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/EPOCHS!){
	    $mode =1; # found SOLUTION/EPOCH
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_Sol_Epoch():  SOLUTION/EPOCHS not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@sitecode, @pntcode, @SolID, @obscode, @tstart, @tstop, @tmid);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~/^\*/);
	if($line =~m!-SOLUTION/EPOCHS!){  $mode = 2; next; }
	$sitecode[$icnt] = substr($line,1,4);
	$pntcode[$icnt] = substr($line,6,2);
	$SolID[$icnt] = substr($line,9,4);
	$obscode[$icnt] = substr($line,14,1);
	my $yy = substr($line,16,2); 	$yy += $yy < 50 ? 2000 : 1900;
	my $doy = substr($line,19,3);
	my $sec = substr($line,23,5);
	$tstart[$icnt] = new Time_lib;
	$tstart[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$tstart[$icnt] ->pluseq($sec);
	
	$yy = substr($line,29,2); 	$yy += $yy < 50 ? 2000 : 1900;
	$doy = substr($line,32,3);
	$sec = substr($line,36,5);
	$tstop[$icnt] = new Time_lib;
	$tstop[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$tstop[$icnt] ->pluseq($sec);

	$yy = substr($line,42,2); 	$yy += $yy < 50 ? 2000 : 1900;
	$doy = substr($line,45,3);
	$sec = substr($line,49,5);
	$tmid[$icnt] = new Time_lib;
	$tmid[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$tmid[$icnt] ->pluseq($sec);
	
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    my %SolEpoch= (
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'ObsCode', \@obscode,
	'StartTime', \@tstart,
	'StopTime', \@tstop,
	'MidTime', \@tmid,
	'Ndat', $icnt);
    $pkg->{SolutionEpoch} = \%SolEpoch;
    close(IN);
    return $icnt;
}

sub SolIDs($$;$){
    my $pkg = shift;
    my $sitecode1 = shift; # SiteCode
    my $epoch = shift; # 
    $SolEpoch = $pkg->{SolutionEpoch};

    my @solidarr;
    my $ndat= $SolEpoch->{Ndat};
    my $sitecode= $SolEpoch->{SiteCode};
    my $solid= $SolEpoch->{SolutionID};
    my $start= $SolEpoch->{StartTime};
    my $stop= $SolEpoch->{StopTime};
    for( my $i=0; $i < $ndat; $i++){
	next unless($$sitecode[$i] eq $sitecode1);
	if(defined($epoch)){
	    push( @solidarr,($$solid[$i]) )
		if( $$start[$i] <= $epoch && $$stop[$i] >= $epoch) ;
	}else{
	    push(@solidarr,($$solid[$i]));
	}
    }
    return @solidarr;
}

# Solution Epoch
sub SolEpoch($$$){
    my $pkg = shift;
    my $sitecode1 = shift; # SiteCode
    my $solid1 = shift; # Solution ID
    
    $SolEpoch = $pkg->{SolutionEpoch};

    my $ndat= $SolEpoch->{Ndat};
    my $sitecode= $SolEpoch->{SiteCode};
    my $solid= $SolEpoch->{SolutionID};
    my $st= $SolEpoch->{StartTime};
    my $et= $SolEpoch->{StopTime};
    
    my @stet;
    for( my $i=0; $i<$ndat; $i++){
	if($$sitecode[$i] eq $sitecode1 &&
	   $$solid[$i] == $solid1 ){
	    $stet[0] = $$st[$i];
	    $stet[1] = $$et[$i];
	}
    }
    
    if( defined( $stet[0] ) ){
	return @stet;
    }else{
	return ();
    }

}

sub Get_Sol_Est($$){
    my $pkg = shift;
    my $infile = shift;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/ESTIMATE!){
	    $mode =1; # found SOLUTION/ESTIMATE
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_XYZ():  SOLUTION/ESTIMATE not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@prmindx, @prmtype, @sitecode, @pntcode, @SolID, @epoch, @unit, @constr, @prmval, @stdev);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SOLUTION/ESTIMATE!){  $mode = 2; next; }
	$prmindx[$icnt] = substr($line,1,5);
	$prmtype[$icnt] = substr($line,7,6);
	$sitecode[$icnt] = substr($line,14,4);
	$pntcode[$icnt] = substr($line,19,2);
	$SolID[$icnt] = substr($line,22,4);
	my $yy = substr($line,27,2); 	$yy += $yy < 50 ? 2000 : 1900;
	my $doy = substr($line,30,3);
	my $sec = substr($line,34,5);
	$epoch[$icnt] = new Time_lib;
	$epoch[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$epoch[$icnt] ->pluseq($sec);
	
	$unit[$icnt] = substr($line,40,4);
	$constr[$icnt] = substr($line,45,1);
	$prmval[$icnt] = substr($line,47,21);
	$stdev[$icnt] = substr($line,69,11);
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    # my $solest = $pkg->{SolutionEstimate};
    # $solest->{ParamIndex} = \@prmindx;
    # $solest->{ParamType}= \@prmtype;
    # $solest->{SiteCode}= \@sitecode;
    # $solest->{PointCode} =  \@pntcode;
    # $solest->{SolutionID} =\@SolID;
    # $solest->{Epoch} =  \@epoch;
    # $solest->{Unit} = \@unit;
    # $solest->{Constraint} = \@constr;
    # $solest->{ParamVal} = \@prmval;
    # $solest->{ParamSTD} = \@stdev;
    # $solest->{Nparam} = $icnt;
    # $pkg->{SolutionEstimate} = $solest;
    my %SolEst= (
	'ParamIndex', \@prmindx,
	'ParamType', \@prmtype,
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'Epoch', \@epoch,
	'Unit', \@unit,
	'Constraint', \@constr,
	'ParamVal', \@prmval,
	'ParamSTD', \@stdev,
	'Nparam', $icnt );
    $pkg->{SolutionEstimate} = \%SolEst;

    close(IN);
    return $icnt;

}

sub XYZ($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $solid = shift;
    
    my $solest = $pkg->{SolutionEstimate};
    unless(defined($solest)){
	print STDERR "E Sinex::XYZ():  [SolutionEstimate] etnry was not in memory. Get_Sol_Est() first.\n";
	return ();
    }
    
    my $sitcodes = $solest->{SiteCode};
    my $solids = $solest->{SolutionID};
    my $prmvals = $solest->{ParamVal};
    my $prmtypes = $solest->{ParamType};
    my $prmstds = $solest->{ParamSTD};
    my $epochs = $solest->{Epoch};
    my $nprm =  $solest->{Nparam};
    my (@xyz, $epoch);
    for(my $i=0; $i<$nprm; $i++){
	if( $$sitcodes[$i] =~/$stcode/ &&  $$solids[$i]  =~ /$solid/ ){
	    if($$prmtypes[$i] =~/STAX/ ) {$xyz[0] = $$prmvals[$i]; $xyz[3] = $$prmstds[$i];}
	    elsif($$prmtypes[$i] =~/STAY/ ) {$xyz[1] = $$prmvals[$i]; $xyz[4] = $$prmstds[$i];}
	    elsif($$prmtypes[$i] =~/STAZ/ ) {$xyz[2] = $$prmvals[$i]; $xyz[5] = $$prmstds[$i];}
	    $epoch = $$epochs[$i];
	}
    }    

    return (@xyz,$epoch);
}

sub VXYZ($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $solid = shift;
    
    my $solest = $pkg->{SolutionEstimate};
    unless(defined($solest)){
	print STDERR "E Sinex::XYZ():  [SolutionEstimate] etnry was not in memory. Get_Sol_Est() first.\n";
	return ();
    }
    
    my $sitcodes = $solest->{SiteCode};
    my $solids = $solest->{SolutionID};
    my $prmvals = $solest->{ParamVal};
    my $prmtypes = $solest->{ParamType};
    my $prmstds = $solest->{ParamSTD};
    my $epochs = $solest->{Epoch};
    my $nprm =  $solest->{Nparam};
    my (@xyz, $epoch);
    for(my $i=0; $i<$nprm; $i++){
	if( $$sitcodes[$i] =~/$stcode/ &&  $$solids[$i]  =~ /$solid/ ){
	    if($$prmtypes[$i] =~/VELX/ ) {$xyz[0] = $$prmvals[$i]; $xyz[3] = $$prmstds[$i];}
	    elsif($$prmtypes[$i] =~/VELY/ ) {$xyz[1] = $$prmvals[$i]; $xyz[4] = $$prmstds[$i];}
	    elsif($$prmtypes[$i] =~/VELZ/ ) {$xyz[2] = $$prmvals[$i]; $xyz[5] = $$prmstds[$i];}
	    $epoch = $$epochs[$i];
	}
    }    

    return (@xyz,$epoch);
}


sub Get_PSD($$){
    my $pkg = shift;
    my $infile = shift;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open PSD sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/ESTIMATE!){
	    $mode =1; # found SOLUTION/ESTIMATE
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_Psd():  SOLUTION/ESTIMATE not found \n";
	return -1;
    }

    my $icnt=0;
    my $line = <IN>;
    my (@prmindx, @prmtype, @sitecode, @pntcode, @SolID, @epoch, @unit, @constr, @prmval, @stdev);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SOLUTION/ESTIMATE!){  $mode = 2; next; }
	$prmindx[$icnt] = substr($line,1,5);
	$prmtype[$icnt] = substr($line,7,6);
	$sitecode[$icnt] = substr($line,14,4);
	$pntcode[$icnt] = substr($line,19,2);
	$SolID[$icnt] = substr($line,22,4);
	my $yy = substr($line,27,2); 	$yy += $yy < 50 ? 2000 : 1900;
	my $doy = substr($line,30,3);
	my $sec = substr($line,34,5);
	$epoch[$icnt] = new Time_lib;
	$epoch[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$epoch[$icnt] ->pluseq($sec);
	
	$unit[$icnt] = substr($line,40,4);
	$constr[$icnt] = substr($line,45,1);
	$prmval[$icnt] = substr($line,47,21);
	$stdev[$icnt] = substr($line,69,11);
	$icnt++;
    }continue{
	$line =<IN>;
    }
    

    my %SolEst= (
	'ParamIndex', \@prmindx,
	'ParamType', \@prmtype,
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'Epoch', \@epoch,
	'Unit', \@unit,
	'Constraint', \@constr,
	'ParamVal', \@prmval,
	'ParamSTD', \@stdev,
	'Nparam', $icnt );
    $pkg->{SolutionPSD} = \%SolEst;
    close(IN);
    return $icnt;

}

## Post Sisemic Deformation model in ITRF2014 
sub dL_psd($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $epoch = shift;
    
    my $solpsd = $pkg->{SolutionPSD};
    unless(defined($solpsd)){
	print STDERR "E Sinex::PSD():  [SolutionEstimate] etnry was not in memory. Get_PSD() first.\n";
	return ();
    }

    my $sitcodes = $solpsd->{SiteCode};
    my $solids = $solpsd->{SolutionID};
    my $prmvals = $solpsd->{ParamVal};
    my $prmtypes = $solpsd->{ParamType};
    my $prmstds = $solpsd->{ParamSTD};
    my $epochs = $solpsd->{Epoch};
    my $nprm =  $solpsd->{Nparam};

    my (@exp_enh, @log_enh);
    my @iexp=(-1,-1,-1);
    my @ilog=(-1,-1,-1);
    for(my $i=0; $i<$nprm; $i++){
	if( $$sitcodes[$i] =~/$stcode/ ){
	    next if($epoch < $$epochs[$i]);
	    if($$prmtypes[$i] =~/AEXP_E/ ) {
		$iexp[0]++;
		my @earr=(); $exp_enh[$iexp[0]] = \@earr;
		$exp_enh[$iexp[0]][0] = $$prmvals[$i]; 
		$exp_enh[$iexp[0]][1] = $$prmstds[$i];
		$exp_enh[$iexp[0]][2] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/AEXP_N/ ) {
		$iexp[1]++;
		my @earr=(); $exp_enh[$iexp[1]] = \@earr;
		$exp_enh[$iexp[1]][0] = $$prmvals[$i]; 
		$exp_enh[$iexp[1]][1] = $$prmstds[$i];
		$exp_enh[$iexp[1]][2] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/AEXP_H/ ) {
		$iexp[2]++;
		my @earr=(); $exp_enh[$iexp[2]] = \@earr;
		$exp_enh[$iexp[2]][0] = $$prmvals[$i]; 
		$exp_enh[$iexp[2]][1] = $$prmstds[$i];
		$exp_enh[$iexp[2]][2] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/TEXP_E/ ) {
		$exp_enh[$iexp[0]][3] = $$prmvals[$i]; 
		$exp_enh[$iexp[0]][4] = $$prmstds[$i];
		$exp_enh[$iexp[0]][5] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/TEXP_N/ ) {
		$exp_enh[$iexp[1]][3] = $$prmvals[$i]; 
		$exp_enh[$iexp[1]][4] = $$prmstds[$i];
		$exp_enh[$iexp[1]][5] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/TEXP_H/ ) {
		$exp_enh[$iexp[2]][3] = $$prmvals[$i]; 
		$exp_enh[$iexp[2]][4] = $$prmstds[$i];
		$exp_enh[$iexp[2]][5] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/ALOG_E/ ) {
		$ilog[0]++;
		my @earr=(); $log_enh[$ilog[0]] = \@earr;
		$log_enh[$ilog[0]][0] = $$prmvals[$i]; 
		$log_enh[$ilog[0]][1] = $$prmstds[$i];
		$log_enh[$ilog[0]][2] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/ALOG_N/ ) {
		$ilog[1]++;
		my @earr=(); $log_enh[$ilog[1]] = \@earr;
		$log_enh[$ilog[1]][0] = $$prmvals[$i]; 
		$log_enh[$ilog[1]][1] = $$prmstds[$i];
		$log_enh[$ilog[1]][2] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/ALOG_H/ ) {
		$ilog[2]++;
		my @earr=(); $log_enh[$ilog[2]] = \@earr;
		$log_enh[$ilog[2]][0] = $$prmvals[$i]; 
		$log_enh[$ilog[2]][1] = $$prmstds[$i];
		$log_enh[$ilog[2]][2] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/TLOG_E/ ) {
		$log_enh[$ilog[0]][3] = $$prmvals[$i]; 
		$log_enh[$ilog[0]][4] = $$prmstds[$i];
		$log_enh[$ilog[0]][5] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/TLOG_N/ ) {
		$log_enh[$ilog[1]][3] = $$prmvals[$i]; 
		$log_enh[$ilog[1]][4] = $$prmstds[$i];
		$log_enh[$ilog[1]][5] = $$epochs[$i];
	    }elsif($$prmtypes[$i] =~/TLOG_H/ ) {
		$log_enh[$ilog[2]][3] = $$prmvals[$i]; 
		$log_enh[$ilog[2]][4] = $$prmstds[$i];
		$log_enh[$ilog[2]][5] = $$epochs[$i];
	    }

	}
    }    

    my @denh = (0,0,0);
    for($i=0; $i<3; $i++){ # for ENH
	for($j=0; $j<$iexp[$i]+1; $j++){ # Exp 
	    my $dt = $epoch - $exp_enh[$j][5];
	    my $dy = $dt/86400.0/365.25;
	    $denh[$i] += &exp_psd($exp_enh[$j][0], $dy, $exp_enh[$j][3] );
	}
	for($j=0; $j<$ilog[$i]+1; $j++){ # log
	    my $dt = $epoch - $log_enh[$j][5];
	    my $dy = $dt/86400.0/365.25;
	    $denh[$i] += &log_psd($log_enh[$j][0], $dy, $log_enh[$j][3] );
	}

    }

    my @xyz0 = $pkg->XYZ_at($stcode,$epoch);
#    my @xyz0 = $pkg->XYZ_fwd($stcode,$epoch);
    my $xyz1 = new Math::Matrix(
		[$xyz0[0],0,0],
		[$xyz0[1],0,0],
		[$xyz0[2],0,0] );
    my $denh1 = new Math::Matrix(
	[$denh[0], 0,0],
	[$denh[1], 0,0],
	[$denh[2], 0,0] );
    my $geo = new GeoCord;
    my $dxyz1 = $geo->ENH2XYZ($xyz1,$denh1);
    my @dxyz2 = ($$dxyz1[0][0],$$dxyz1[1][0],$$dxyz1[2][0]);
    return (@denh,  @dxyz2);
}

sub exp_psd( $$$ ){
    my $a = shift;
    my $dt = shift;
    my $tau = shift;

    my $qte = - $dt/$tau;
    my $ret = $a*(1.0 - exp($qte));
    return $ret;
}

sub log_psd( $$$ ){
    my $a = shift;
    my $dt = shift;
    my $tau = shift;

    my $qte = $dt / $tau ;
    my $ret = $a * log(1.0 + $qte);
    return $ret;
}



# 
#  Sitecode, Epoch,; SolID
# sub XYZ_at( $$$;$ ){
#     my $pkg = shift;
#     my $stcode = shift;
#     my $epoch  = shift;
#     my $solid    = shift;
    
#     unless(defined($solid)){
# 	my @solids = $pkg->SolIDs($stcode,$epoch);    
#     	for(my $i=0; $i<@solids; $i++){
#     	    my @stet = $pkg->SolEpoch($stcode,$solids[$i]);
#     	    if($epoch >= $stet[0] && $epoch <= $stet[1]){
#     		$solid = $solids[$i]; last;
#     	    }
#     	} 
#     }

#     for(my $i=0; $i<@solids; $i++){
# 	my @stet = $pkg->SolEpoch($stcode,$solids[$i]);
# 	if($epoch >= $stet[0] && $epoch <= $stet[1]){
# 	    $solid = $solids[$i]; last;
# 	}
#     } 

#     unless(defined($solid)){
# 	printf STDERR "E Sinex::XYZ_at(): Failed to find epoch (%s) of sitecode(%d) \n",
# 	$epoch->Str_fmt1(1), $stecode;
# 	return ();
#     }

#     my @xyz = $pkg->XYZ($stcode, $solid );
#     my @vxyz = $pkg->VXYZ($stcode, $solid );
#     my $pos_epoch = pop(@xyz);    
#     my $vel_epoch = pop(@vxyz);    
#     my $dt = $epoch - $pos_epoch ; # sec
#     my $dy = $dt / 86400.0 / 365.25 ; # yr
#     my @dxyz = map( $_ * $dy , @vxyz );
#     my @xyz1 = map(  $xyz[$_] + $dxyz[$_] , (0,1,2) );
#     my @ve = map( sqrt( $dxyz[$_] *$dxyz[$_] + $xyz[$_] *$xyz[$_] ), (3,4,5) );
#     return (@xyz1, @ve);
# }

sub XYZ_at( $$$ ){
    my $pkg = shift;
    my $stcode = shift;
    my $epoch  = shift;
    my $solid;
    
    my @solids = $pkg->SolIDs($stcode,$epoch);    

    my (@xyz1, @ve);
    for(my $i=0; $i<@solids; $i++){
	my @xyz = $pkg->XYZ($stcode, $solids[$i] );
	my @vxyz = $pkg->VXYZ($stcode, $solids[$i] );
	my $pos_epoch = pop(@xyz);    
	my $vel_epoch = pop(@vxyz);    
	next if($epoch < $pos_epoch);
	my $dt = $epoch - $pos_epoch ; # sec
	my $dy = $dt / 86400.0 / 365.25 ; # yr
	my @dxyz = map( $_ * $dy , @vxyz );
	@xyz1 = map(  $xyz[$_] + $dxyz[$_] , (0,1,2) );
	@ve = map( sqrt( $dxyz[$_] *$dxyz[$_] + $xyz[$_] *$xyz[$_] ), (3,4,5) );
    }

    if(defined($solpsd)){
	my (@denh, @dxyz) = dL_psd($stcode,$epoch);
	my @xyz2 = @xyz1;
	@xyz1 = map(  $xyz2[$_] + $dxyz[$_] , (0,1,2) );
	
	my $epoch1 = $epoch + (-10.*86400.);
	my $epoch2 = $epoch + 10.*86400.;
	my (@denh1, @dxyz1) = dL_psd($stcode,$epoch1);
	my (@denh2, @dxyz2) = dL_psd($stcode,$epoch2);
	my @ve2= map( ($dxyz2[$_] - $dxyz1[$_])/20.*365.25, (0,1,2 );

	my @ve1 = @ve;
	@ve = map(  $ve1[$_] + $ve2[$_] , (0,1,2) );

    }

    return (@xyz1, @ve);

}

# 
#  Sitecode, Epoch,; SolID
# forward extrapolation of Position if not within the epoch range.
# sub XYZ_fwd( $$$;$ ){
#     my $pkg = shift;
#     my $stcode = shift;
#     my $epoch  = shift;
#     my $solid    = shift;
        
#     unless(defined($solid)){
# #	my @solids = $pkg->SolIDs($stcode,$epoch);
# 	my @solids = $pkg->SolIDs($stcode);
# 	for(my $i=0; $i<@solids; $i++){
# 	    my @stet = $pkg->SolEpoch($stcode,$solids[$i]);
# 	    if($epoch >= $stet[0]){$solid = $solids[$i];}
# 	    if( $epoch <= $stet[1]){ last;}
# 	} 
#     }
#     unless(defined($solid)){
# 	printf STDERR "E Sinex::XYZ_at(): Requested epoch (%s) for sitecode(%d) was before epochs: \n",
# 	$epoch->Str_fmt1(1), $stcode;
# 	for(my $i=0; $i<@solids; $i++){
# 	    my @stet = $pkg->SolEpoch($stcode,$solids[$i]);
# 	    printf STDERR "%s - %s, ", $stet[0]->Str_fmt1(1),$stet[1]->Str_fmt1(1);
# 	}
# 	print STDERR "\n";
# 	return ();
#     }

#     my @xyz = $pkg->XYZ($stcode, $solid );
#     my @vxyz = $pkg->VXYZ($stcode, $solid );
#     my $pos_epoch = pop(@xyz);    
#     my $vel_epoch = pop(@vxyz);    
#     my $dt = $epoch - $pos_epoch ; # sec
#     my $dy = $dt / 86400.0 / 365.25 ; # yr
#     my @dxyz = map( $_ * $dy , @vxyz );
#     my @xyz1 = map(  $xyz[$_] + $dxyz[$_] , (0,1,2) );
#     my @ve = map( sqrt( $dxyz[$_] *$dxyz[$_] + $xyz[$_] *$xyz[$_] ), (3,4,5) );
#     return (@xyz1, @ve);
# }


sub Epoch2SolID($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $epoch = shift;
    
    my @solids = $pkg->SolIDs($stcode,$epoch);
    
    return @solids;
 }

1;



