# @Version: $Id: Sinexx.pm 81 2016-01-21 10:56:01Z vlbi $
#
package Sinexx;
#use Exporter;
#use Carp;
use Time_lib;
use GeoCord;

$dbg =0;
BEGIN{

}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    return $this;
}

sub Read_Sinex($$){
    my $pkg=shift;
    my $infile=shift;
    my $n1 = $pkg->Get_Site_Id($infile);
    my $n2 = $pkg->Get_Sol_Epoch($infile);
    my $n3 = $pkg->Get_Sol_Est($infile);
}

sub Read_Sinex_ext($$){
    my $pkg=shift;
    my $infile=shift;
    my $n1 = $pkg->Get_Site_Id($infile);
    my $n2 = $pkg->Get_Sol_Epoch($infile);
    my $n3 = $pkg->Get_Sol_Est_Ext($infile);
}

sub Get_Site_Id( $$ ){
    my ($pkg, $infile) = @_;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SITE/ID!){
	    $mode =1; # found SITE/ID
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_Site_Id():  SITE/ID not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@sitecode, @pntcode, @MonuID, @obscode, @sitedesc, @lon, @lat, @height);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SITE/ID!){  $mode = 2; next; }
	$sitecode[$icnt] = substr($line,1,4);
	$pntcode[$icnt] = substr($line,6,2);
	$MonuID[$icnt] = substr($line,9,9);
	$obscode[$icnt] = substr($line,19,1);
	$sitedesc[$icnt] = substr($line,21,22);
	my $l1 = substr($line,44,3);
	my $sg = $l1 > 0 ? 1 : -1;  $l1 *= $sg;
	my $l2 = substr($line,48,2);
	my $l3 = substr($line,51,4);
	$lon[$icnt] = ($l3/60. + $l2)/60. + $l1; 
	$lon[$icnt] *= $sg;

	$l1 = substr($line,56,3);
	$sg = $l1 > 0 ? 1 : -1;  $l1 *= $sg;
	$l2 = substr($line,60,2);
	$l3 = substr($line,63,4);
	$lat[$icnt] = ($l3/60. + $l2)/60. + $l1; 
	$lat[$icnt] *= $sg;
	
	$height[$icnt] = substr($line,68,7);
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    my %siteid= (
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'MonumentID', \@MonuID,
	'ObsCode', \@obscode,
	'SiteDescr', \@sitedesc,
	'Longitude', \@lon,
	'Latitude', \@lat,
	'Height', \@height,
	'Nsite', $icnt);
    $pkg->{siteid} = \%siteid;
    close(IN);
    return $icnt;
}

# Number of Sites
sub Nsite($){
    my $pkg= shift;
    $siteid = $pkg->{siteid};
    return $siteid->{Nsite};
}

sub SiteDescs($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{SiteDescr};
}

# Return Site-Description of the sitecode
sub SiteDesc($$){
    my $pkg = shift;
    my $stcode = shift;
    my $sitedesc = $pkg->{siteid}->{SiteDescr};
    my $sitecode = $pkg->{siteid}->{SiteCode};
    for(my $i=0; $i<@$sitecode; $i++){
	if($$sitecode[$i] =~/$stcode/ ){
	    return $$sitedesc[$i];
	}
    }
    return ""; 
}

sub SiteCodes($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{SiteCode};
}

## Find pattern in the site description
sub SiteCodes_DescPtn($$){
    my $pkg = shift;
    my $ptn = shift;
    $ptn = lc($ptn);

    my $siteid = $pkg->{siteid};
    my $siteDesc = $siteid->{SiteDescr};
    my $siteCode = $siteid->{SiteCode};
    my $found =0;
    my @ids =();
    for(my $i=0; $i< @$siteDesc; $i++){
	if( lc($$siteDesc[$i]) =~/$ptn/ ){
	    $ids[$found] = $$siteCode[$i] ;
#	    push(@ids, ( $$siteCode[$i]));
	    $found++;
	}
    }
    return @ids;
}


# Longitude
sub SiteLon($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{Longitude};
}

# Latitude
sub SiteLat($){
    my $pkg = shift;
    $siteid = $pkg->{siteid};
    return $siteid->{Latitude};
}


sub Get_Sol_Epoch( $$ ){
    my ($pkg, $infile) = @_;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/EPOCHS!){
	    $mode =1; # found SOLUTION/EPOCH
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_Sol_Epoch():  SOLUTION/EPOCHS not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@sitecode, @pntcode, @SolID, @obscode, @tstart, @tstop, @tmid);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~/^\*/);
	if($line =~m!-SOLUTION/EPOCHS!){  $mode = 2; next; }
	$sitecode[$icnt] = substr($line,1,4);
	$pntcode[$icnt] = substr($line,6,2);
	$SolID[$icnt] = substr($line,9,4);
	$obscode[$icnt] = substr($line,14,1);
	my $yy = substr($line,16,2); 	$yy += $yy < 50 ? 2000 : 1900;
	my $doy = substr($line,19,3);
	my $sec = substr($line,23,5);
	$tstart[$icnt] = new Time_lib;
	$tstart[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$tstart[$icnt] ->pluseq($sec);
	
	$yy = substr($line,29,2); 	$yy += $yy < 50 ? 2000 : 1900;
	$doy = substr($line,32,3);
	$sec = substr($line,36,5);
	$tstop[$icnt] = new Time_lib;
	$tstop[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$tstop[$icnt] ->pluseq($sec);

	$yy = substr($line,42,2); 	$yy += $yy < 50 ? 2000 : 1900;
	$doy = substr($line,45,3);
	$sec = substr($line,49,5);
	$tmid[$icnt] = new Time_lib;
	$tmid[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$tmid[$icnt] ->pluseq($sec);
	
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    my %SolEpoch= (
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'ObsCode', \@obscode,
	'StartTime', \@tstart,
	'StopTime', \@tstop,
	'MidTime', \@tmid,
	'Ndat', $icnt);
    $pkg->{SolutionEpoch} = \%SolEpoch;
    close(IN);
    return $icnt;
}

# Return Solution id array
# argument1: Sitecode:
sub SolIDs($$;$){
    my $pkg = shift;
    my $sitecode1 = shift; # SiteCode
    my $epoch = shift; # SiteCode
    $SolEpoch = $pkg->{SolutionEpoch};

    my @solidarr;
    my $ndat= $SolEpoch->{Ndat};
    my $sitecode= $SolEpoch->{SiteCode};
    my $solid= $SolEpoch->{SolutionID};
    my $start= $SolEpoch->{StartTime};
    my $stop= $SolEpoch->{StopTime};
    for( my $i=0; $i < $ndat; $i++){
	next unless($$sitecode[$i] eq $sitecode1);
	if(defined($epoch)){
	    push( @solidarr,($$solid[$i]) )
		if( $$start[$i] <= $epoch && $$stop[$i] >= $epoch) ;
	}else{
	    push(@solidarr,($$solid[$i]));
	}
    }
    return @solidarr;
}


## Finding Solutioin id In Solution/Estimate_ext epochs
# Return Solution ID
# argument 1: sitecode
# argument2: Epoch
sub EstSolIDs($$;$){
    my $pkg = shift;
    my $sitecode1 = shift; # SiteCode
    my $epoch = shift; # Epoch
    $SolEpoch = $pkg->{SolutionEpoch};

    my @solidarr;
    my $ndat= $SolEpoch->{Ndat};
    my $sitecode= $SolEpoch->{SiteCode};
    my $solid= $SolEpoch->{SolutionID};
    my $start= $SolEpoch->{StartTime};
    my $stop= $SolEpoch->{StopTime};
    for( my $i=0; $i < $ndat; $i++){
	next unless($$sitecode[$i] eq $sitecode1);
	if(defined($epoch)){
	    push( @solidarr,($$solid[$i]) )
		if( $$start[$i] <= $epoch && $$stop[$i] >= $epoch) ;
	}else{
	    push(@solidarr,($$solid[$i]));
	}
    }
    return @solidarr;
}


# Solution Epoch
sub SolEpoch($$$){
    my $pkg = shift;
    my $sitecode1 = shift; # SiteCode
    my $solid1 = shift; # Solution ID
    
    $SolEpoch = $pkg->{SolutionEpoch};

    my $ndat= $SolEpoch->{Ndat};
    my $sitecode= $SolEpoch->{SiteCode};
    my $solid= $SolEpoch->{SolutionID};
    my $st= $SolEpoch->{StartTime};
    my $et= $SolEpoch->{StopTime};
    
    my @stet;
    for( my $i=0; $i<$ndat; $i++){
	if($$sitecode[$i] eq $sitecode1 &&
	   $$solid[$i] == $solid1 ){
	    $stet[0] = $$st[$i];
	    $stet[1] = $$et[$i];
	}
    }
    
    if( defined( $stet[0] ) ){
	return @stet;
    }else{
	return ();
    }

}

sub Get_Sol_Est($$){
    my $pkg = shift;
    my $infile = shift;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/ESTIMATE!){
	    $mode =1; # found SOLUTION/ESTIMATE
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_XYZ():  SOLUTION/ESTIMATE not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@prmindx, @prmtype, @sitecode, @pntcode, @SolID, @epoch, @unit, @constr, @prmval, @stdev);
    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SOLUTION/ESTIMATE!){  $mode = 2; next; }
	$prmindx[$icnt] = substr($line,1,5);
	$prmtype[$icnt] = substr($line,7,6);
	$sitecode[$icnt] = substr($line,14,4);
	$pntcode[$icnt] = substr($line,19,2);
	$SolID[$icnt] = substr($line,22,4);
	my $yy = substr($line,27,2); 	$yy += $yy < 50 ? 2000 : 1900;
	my $doy = substr($line,30,3);
	my $sec = substr($line,34,5);
	$epoch[$icnt] = new Time_lib;
	$epoch[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$epoch[$icnt] ->pluseq($sec);
	
	$unit[$icnt] = substr($line,40,4);
	$constr[$icnt] = substr($line,45,1);
	$prmval[$icnt] = substr($line,47,21);
	$stdev[$icnt] = substr($line,69,11);
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    # my $solest = $pkg->{SolutionEstimate};
    # $solest->{ParamIndex} = \@prmindx;
    # $solest->{ParamType}= \@prmtype;
    # $solest->{SiteCode}= \@sitecode;
    # $solest->{PointCode} =  \@pntcode;
    # $solest->{SolutionID} =\@SolID;
    # $solest->{Epoch} =  \@epoch;
    # $solest->{Unit} = \@unit;
    # $solest->{Constraint} = \@constr;
    # $solest->{ParamVal} = \@prmval;
    # $solest->{ParamSTD} = \@stdev;
    # $solest->{Nparam} = $icnt;
    # $pkg->{SolutionEstimate} = $solest;
    my %SolEst= (
	'ParamIndex', \@prmindx,
	'ParamType', \@prmtype,
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'Epoch', \@epoch,
	'Unit', \@unit,
	'Constraint', \@constr,
	'ParamVal', \@prmval,
	'ParamSTD', \@stdev,
	'Nparam', $icnt );
    $pkg->{SolutionEstimate} = \%SolEst;

    close(IN);
    return $icnt;

}


sub XYZ($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $solid = shift;
    
    my @indx = $pkg->Find_ParamIndx($stcode);
    for(my $i=0; $i<@indx; $i++){
	
    }

    my $solest = $pkg->{SolutionEstimate};
    unless(defined($solest)){
	print STDERR "E Sinex::XYZ():  [SolutionEstimate] etnry was not in memory. Get_Sol_Est() first.\n";
	return ();
    }
    
    my $sitcodes = $solest->{SiteCode};
    my $solids = $solest->{SolutionID};
    my $prmvals = $solest->{ParamVal};
    my $prmtypes = $solest->{ParamType};
    my $prmstds = $solest->{ParamSTD};
    my $epochs = $solest->{Epoch};
    my $nprm =  $solest->{Nparam};
    my (@xyz, $epoch);
    for(my $i=0; $i<$nprm; $i++){
	if( $$sitcodes[$i] =~/$stcode/){
	    if( $$solids[$i]  =~ /$solid/ ){
		if($$prmtypes[$i] =~/STAX/ ) {$xyz[0] = $$prmvals[$i]; $xyz[3] = $$prmstds[$i];}
		elsif($$prmtypes[$i] =~/STAY/ ) {$xyz[1] = $$prmvals[$i]; $xyz[4] = $$prmstds[$i];}
		elsif($$prmtypes[$i] =~/STAZ/ ) {$xyz[2] = $$prmvals[$i]; $xyz[5] = $$prmstds[$i];}
		$epoch = $$epochs[$i];
	    }
	}
    }    

    return (@xyz,$epoch);
}

sub VXYZ($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $solid = shift;
    
    my $solest = $pkg->{SolutionEstimate};
    unless(defined($solest)){
	print STDERR "E Sinex::XYZ():  [SolutionEstimate] etnry was not in memory. Get_Sol_Est() first.\n";
	return ();
    }
    
    my $sitcodes = $solest->{SiteCode};
    my $solids = $solest->{SolutionID};
    my $prmvals = $solest->{ParamVal};
    my $prmtypes = $solest->{ParamType};
    my $prmstds = $solest->{ParamSTD};
    my $epochs = $solest->{Epoch};
    my $nprm =  $solest->{Nparam};
    my (@xyz, $epoch);
    for(my $i=0; $i<$nprm; $i++){
	if( $$sitcodes[$i] =~/$stcode/ &&  $$solids[$i]  =~ /$solid/ ){
	    if($$prmtypes[$i] =~/VELX/ ) {$xyz[0] = $$prmvals[$i]; 
					  $xyz[3] = $$prmstds[$i];}
	    elsif($$prmtypes[$i] =~/VELY/ ) {$xyz[1] = $$prmvals[$i];
					     $xyz[4] = $$prmstds[$i];}
	    elsif($$prmtypes[$i] =~/VELZ/ ) {$xyz[2] = $$prmvals[$i];
					     $xyz[5] = $$prmstds[$i];}
	    $epoch = $$epochs[$i];
	}
    }    

    return (@xyz,$epoch);
}


sub Get_PSD($$){
    my $pkg = shift;
    my $infile = shift;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open PSD sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/ESTIMATE!){
	    $mode =1; # found SOLUTION/ESTIMATE
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_Psd():  SOLUTION/ESTIMATE not found \n";
	return -1;
    }

    my $icnt=0;
    my $line = <IN>;
    my (@prmindx, @prmtype, @sitecode, @pntcode, @SolID, @epoch, @unit, @constr, @prmval, @stdev);
    my %entry;
    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SOLUTION/ESTIMATE!){  $mode = 2; next; }
	$line=~/\s*(\d+)\s+/;
	$prmindx[$icnt] = $1;
	my $rsd=$';  #';
	chomp( $rsd );
	$entry{$prmindx[$icnt]}=$rsd;
#	$prmindx[$icnt] = substr($line,1,5);
	$prmtype[$icnt] = substr($line,7,6);
	$sitecode[$icnt] = substr($line,14,4);
	$pntcode[$icnt] = substr($line,19,2);
	$SolID[$icnt] = substr($line,22,4);
	my $yy = substr($line,27,2); 	$yy += $yy < 50 ? 2000 : 1900;
	my $doy = substr($line,30,3);
	my $sec = substr($line,34,5);
	$epoch[$icnt] = new Time_lib;
	$epoch[$icnt]->Set_time_ydhms($yy,$doy,0,0,0);
	$epoch[$icnt] ->pluseq($sec);
	
	$unit[$icnt] = substr($line,40,4);
	$constr[$icnt] = substr($line,45,1);
	$prmval[$icnt] = substr($line,47,21);
	$stdev[$icnt] = substr($line,69,11);
	$icnt++;
    }continue{
	$line =<IN>;
    }
    

    my %SolEst= (
	'ParamIndex', \@prmindx,
	'Entry', \%entry,
	'ParamType', \@prmtype,
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'Epoch', \@epoch,
	'Unit', \@unit,
	'Constraint', \@constr,
	'ParamVal', \@prmval,
	'ParamSTD', \@stdev,
	'Nparam', $icnt );
    $pkg->{SolutionPSD} = \%SolEst;
    close(IN);
    return $icnt;

}

# return indexes
# arg1: sitecode
sub PSD_Index($$){
    my $pkg=shift;
    my $sitecode = shift;

    my $psd = $pkg->{SolutionPSD};
    my $ndat = $psd->{Nparam};
    my $sitecodes = $psd->{SiteCode};
    my $indxs =  $psd->{ParamIndex};
    my @ret = ();
    for(my $i=0; $i < $ndat; $i++){
	if($$sitecodes[$i] == $sitecode){
	    push( @ret, ($$indxs[$i]) );
	}
    }

    return @ret;
}

# return Entry
# arg1: index
sub PSD_Entry($$){
    my $pkg=shift;
    my $indx = shift;

    my $psd = $pkg->{SolutionPSD};
    my $entries = $psd->{Entry};
    my $ret = $entries->{$indx};
    return $ret;
}


## Post Sisemic Deformation model in ITRF2014 
sub dL_psd($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $epoch = shift;
    
    my $solpsd = $pkg->{SolutionPSD};
    unless(defined($solpsd)){
	print STDERR "E Sinex::PSD():  [SolutionEstimate] etnry was not in memory. Get_PSD() first.\n";
	return ();
    }

    my $sitcodes = $solpsd->{SiteCode};
    my $solids = $solpsd->{SolutionID};
    my $prmvals = $solpsd->{ParamVal};
    my $prmtypes = $solpsd->{ParamType};
    my $prmstds = $solpsd->{ParamSTD};
    my $epochs = $solpsd->{Epoch};
    my $nprm =  $solpsd->{Nparam};

    my @exp_enh=((),(),()); 
    my @log_enh =((),(),()); 
    my @iexp=(-1,-1,-1);
    my @ilog=(-1,-1,-1);
    for(my $i=0; $i<$nprm; $i++){
	if( $$sitcodes[$i] =~/$stcode/ ){
	    next if($epoch < $$epochs[$i]);
	    if($$prmtypes[$i] =~/AEXP_E/ ) {
		$iexp[0]++;
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$exp_enh[0][$iexp[0]][0] = \@earr;
	    }elsif($$prmtypes[$i] =~/AEXP_N/ ) {
		$iexp[1]++;
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$exp_enh[1][$iexp[1]][0] = \@earr;
	    }elsif($$prmtypes[$i] =~/AEXP_H/ ) {
		$iexp[2]++;
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$exp_enh[2][$iexp[2]][0] = \@earr;
	    }elsif($$prmtypes[$i] =~/TEXP_E/ ) {
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$exp_enh[0][$iexp[0]][1] = \@earr;
	    }elsif($$prmtypes[$i] =~/TEXP_N/ ) {
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$exp_enh[1][$iexp[1]][1] = \@earr;
	    }elsif($$prmtypes[$i] =~/TEXP_H/ ) {
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$exp_enh[2][$iexp[2]][1] = \@earr;
	    }elsif($$prmtypes[$i] =~/ALOG_E/ ) {
		$ilog[0]++;
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$log_enh[0][$ilog[0]][0] = \@earr;
	    }elsif($$prmtypes[$i] =~/ALOG_N/ ) {
		$ilog[1]++;
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$log_enh[1][$ilog[1]][0] = \@earr;
	    }elsif($$prmtypes[$i] =~/ALOG_H/ ) {
		$ilog[2]++;
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$log_enh[2][$ilog[2]][0] = \@earr;
	    }elsif($$prmtypes[$i] =~/TLOG_E/ ) {
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$log_enh[0][$ilog[0]][1] = \@earr;
	    }elsif($$prmtypes[$i] =~/TLOG_N/ ) {
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$log_enh[1][$ilog[1]][1] = \@earr;
	    }elsif($$prmtypes[$i] =~/TLOG_H/ ) {
		my @earr=( $$prmvals[$i], $$prmstds[$i],$$epochs[$i]); 
		$log_enh[2][$ilog[2]][1] = \@earr;
	    }

	}
    }    

    my @denh = (0,0,0);
    for($i=0; $i<3; $i++){ # for ENH
	printf " %d: Exp: ",$i if($dbg);
	for($j=0; $j<$iexp[$i]+1; $j++){ # Exp 
	    my $dt = $epoch - $exp_enh[$i][$j][0][2];
	    my $dy = $dt/86400.0/365.25;
	    my $dx = &exp_psd($exp_enh[$i][$j][0][0], $dy, $exp_enh[$i][$j][1][0] );
	    $denh[$i] += $dx;
	    printf "(%9.2e,%.4f,%.4f )%10.3e ", $exp_enh[$i][$j][0][0], $dy, $exp_enh[$i][$j][1][0] ,$dx if($dbg);
	}
	print "LOG: " if($dbg);
	for($j=0; $j< $ilog[$i]+1; $j++){ # log
	    my $dt = $epoch - $log_enh[$i][$j][0][2];
	    my $dy = $dt/86400.0/365.25;
	    my $dx = &log_psd($log_enh[$i][$j][0][0], $dy, $log_enh[$i][$j][1][0] );
	    $denh[$i] += $dx;
	    printf "%10.3e(%9.2e,%.4f,%.4f )", $dx,$log_enh[$i][$j][0][0], $dy, $log_enh[$i][$j][1][0] if($dbg);;
    }
	print "\n" if($dbg);
    }
    
    my @xyz0 = $pkg->XYZ_at($stcode,$epoch);
    my $xyz1 = new Math::Matrix(
		[$xyz0[0],0,0],
		[$xyz0[1],0,0],
		[$xyz0[2],0,0] );
    my $denh1 = new Math::Matrix(
	[$denh[0], 0,0],
	[$denh[1], 0,0],
	[$denh[2], 0,0] );
    my $geo = new GeoCord;
    my $dxyz1 = $geo->ENH2XYZ($xyz1,$denh1);
    my @dxyz2 = ($$dxyz1[0][0],$$dxyz1[1][0],$$dxyz1[2][0]);
    return (@denh,  @dxyz2);
}

sub exp_psd( $$$ ){
    my $a = shift;
    my $dt = shift;
    my $tau = shift;

    my $qte = - $dt/$tau;
    my $ret = $a*(1.0 - exp($qte));
    return $ret;
}

sub log_psd( $$$ ){
    my $a = shift;
    my $dt = shift;
    my $tau = shift;

    my $qte = $dt / $tau ;
    my $ret = $a * log(1.0 + $qte);
    return $ret;
}



# 
#  Sitecode, Epoch,; SolID
sub XYZ_at( $$$;$ ){
    my $pkg = shift;
    my $sitecode   = shift;
    my $epoch  = shift;
        
    my @xyz=();
    my @xverr =();
    my @indxarr = $pkg->Find_ParamIndx($sitecode,$epoch);
    my ($pepoch, $flg);
    foreach $idx (@indxarr){
	my $entry = $pkg->SolEstExt_Entry($idx);
	my $type = &Ext_Entry_Type($entry);
	if($type eq "STAX"){
	    $xyz[0] = &Ext_Entry($entry,13);
	    $xverr[0] = &Ext_Entry($entry,14);
	    ($pepoch,$flg) = &Ext_RateEpoch($entry);
	}elsif($type eq "STAY"){
	    $xyz[1] = &Ext_Entry($entry,13);
	    $xverr[1] = &Ext_Entry($entry,14);
	}elsif($type eq "STAZ"){
	    $xyz[2] = &Ext_Entry($entry,13);
	    $xverr[2] = &Ext_Entry($entry,14);
	}elsif($type eq "VELX"){
	    $xyz[3] = &Ext_Entry($entry,13);
	    $xverr[3] = &Ext_Entry($entry,14);
	}elsif($type eq "VELY"){
	    $xyz[4] = &Ext_Entry($entry,13);
	    $xverr[4] = &Ext_Entry($entry,14);
	}elsif($type eq "VELZ"){
	    $xyz[5] = &Ext_Entry($entry,13);
	    $xverr[5] = &Ext_Entry($entry,14);
	}
    }
    
    my $dtsec = $epoch - $pepoch; 
    my $dy = $dtsec / ( 86400. * 365.25 ) ;
    my @dxyz = map( $xyz[$_] * $dy, (3,4,5) );
    
    my @dxyz_err = map( $xverr[$_] * $dy, (3,4,5) );
    my @poserr = map( sqrt($xverr[$_]*$xverr[$_] + $dxyz_err[$_] * $dxyz_err[$_]), (0,1,2) );
    for(my $i=0; $i<3; $i++){ $poserr[$i+3] = $xverr[$i+3]; $xyz[$i] += $dxyz[$i]}

    return (@xyz, @poserr);
}

sub Epoch2SolID($$$){
    my $pkg = shift;
    my $stcode = shift;
    my $epoch = shift;
    
    my @solids = $pkg->SolIDs($stcode,$epoch);
    
    return @solids;
 }

# sub Get_Sol_Est_Ext($$){
#     my $pkg = shift;
#     my $infile = shift;
    
#     unless(open(IN, $infile) ){
# 	print  STDERR "E failed to open sinexfile($infile) \n";
# 	return -1;
#     }

#     my $mode=0;
#     while( <IN> ){
# 	if(m!\+SOLUTION/ESTIMATE!){
# 	    $mode =1; # found SOLUTION/ESTIMATE
# 	    last;
# 	}
#     }
#     if($mode ==0){
# 	print STDERR "E Sinex::Get_XYZ():  SOLUTION/ESTIMATE not found \n";
# 	return -1;
#     }
    
#     my $icnt=0;
#     my $line = <IN>;
#     my (@prmindx, @prmtype, @sat1, @sat2,@sitecode, @site2);
#     my ( @pntcode, @SolID, @ref_epoch, @t1_epoch, @t2_epoch, @rate_epoch); 
#     my ( @unit, @constr, @prmval, @stdev );
#     while( !eof(IN) && $mode == 1 ){
# 	next if($line=~ /^\*/);
# 	if($line =~m!-SOLUTION/ESTIMATE_EXT!){  $mode = 2; next; }
# 	$line =~/\s*/; my @terms = split(/\s+/,$'); #');
# 	$prmindx[$icnt] = $terms[0];
# 	$prmtype[$icnt] = $terms[1];
# 	$sat1[$icnt] = $terms[2];
# 	$sat2[$icnt] = $terms[3];
# 	$sitecode[$icnt] = $terms[4];
# 	$site2[$icnt] = $terms[5];
# 	$pntcode[$icnt] = $terms[6];
# 	$SolID[$icnt] = $terms[7];
# 	$ref_epoch[$icnt] = $terms[8];
# 	$t1_epoch[$icnt] = $terms[9];
# 	$t2_epoch[$icnt] = $terms[10];
# 	$rate_epoch[$icnt] = $terms[11];
# 	$unit[$icnt] = $terms[12];
# 	$constr[$icnt] = $terms[13];
# 	$prmval[$icnt] = $terms[14];
# 	$stdev[$icnt] = $terms[15];
# 	$icnt++;
#     }continue{
# 	$line =<IN>;
#     }
    
#     my %SolEst_ext= (
# 	'ParamIndex', \@prmindx,
# 	'ParamType', \@prmtype,
# 	'SiteCode', \@sitecode,
# 	'PointCode', \@pntcode,
# 	'SolutionID', \@SolID,
# 	'Ref_Epoch', \@ref_epoch,
# 	'T1_Epoch', \@t1_epoch,
# 	'T2_Epoch', \@t2_epoch,
# 	'Rate_Epoch', \@rate_epoch,
# 	'Unit', \@unit,
# 	'Constraint', \@constr,
# 	'ParamVal', \@prmval,
# 	'ParamSTD', \@stdev,
# 	'Nparam', $icnt );
#     $pkg->{SolutionEstimate_ext} = \%SolEst_ext;

#     my %SolEst= (
# 	'ParamIndex', \@prmindx,
# 	'ParamType', \@prmtype,
# 	'SiteCode', \@sitecode,
# 	'PointCode', \@pntcode,
# 	'SolutionID', \@SolID,
# 	'Epoch', \@epoch,
# 	'Unit', \@unit,
# 	'Constraint', \@constr,
# 	'ParamVal', \@prmval,
# 	'ParamSTD', \@stdev,
# 	'Nparam', $icnt );
#     $pkg->{SolutionEstimate} = \%SolEst;

#     close(IN);
#     return $icnt;

# }

sub Get_Sol_Est_Ext($$){
    my $pkg = shift;
    my $infile = shift;
    
    unless(open(IN, $infile) ){
	print  STDERR "E failed to open sinexfile($infile) \n";
	return -1;
    }

    my $mode=0;
    while( <IN> ){
	if(m!\+SOLUTION/ESTIMATE_EXT!){
	    $mode =1; # found SOLUTION/ESTIMATE_EXT
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E Sinex::Get_XYZ():  SOLUTION/ESTIMATE_EXT not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <IN>;
    my (@prmindx, %entries );
    my ( @sat1, @sat2,@sitecode, @site2);
    my ( @pntcode, @SolID, @ref_epoch, @t1_epoch, @t2_epoch, @rate_epoch); 
     my ( @unit, @constr, @prmval, @stdev );

    while( !eof(IN) && $mode == 1 ){
	next if($line=~ /^\*/);
	if($line =~m!-SOLUTION/ESTIMATE_EXT!){  $mode = 2; next; }
	unless($line =~/\s*(\d+)\s+/){
	    print STDERR "E Sinexx::Get_Sol_Est_Ext(): Unexpected SOLUTION/ESTIMATE_EXT entry: $line \n";
	    next;
	} 
	$prmindx[$icnt] = $1; 
	my $rsd = $'; # ';
	chomp($rsd);
	$entries{$prmindx[$icnt]} = $rsd; 
	my @terms = split(/\s+/, $rsd);
	$prmtype[$icnt] = $terms[0];
	$sat1[$icnt] = $terms[1];
	$sat2[$icnt] = $terms[2];
	$sitecode[$icnt] = $terms[3];
	$site2[$icnt] = $terms[4];
	$pntcode[$icnt] = $terms[5];
	$SolID[$icnt] = $terms[6];
	$ref_epoch[$icnt] = $terms[7];
	$t1_epoch[$icnt] = $terms[8];
	$t2_epoch[$icnt] = $terms[9];
	$rate_epoch[$icnt] = $terms[10];
	$unit[$icnt] = $terms[11];
	$constr[$icnt] = $terms[12];
	$prmval[$icnt] = $terms[13];
	$stdev[$icnt] = $terms[14];
	$icnt++;
    }continue{
	$line =<IN>;
    }
    
    my %SolEst_ext= (
	'ParamIndex', \@prmindx,
	'Entry', \%entries,
	'SiteCode', \@sitecode,
	'PointCode', \@pntcode,
	'SolutionID', \@SolID,
	'Ref_Epoch', \@ref_epoch,
	'T1_Epoch', \@t1_epoch,
	'T2_Epoch', \@t2_epoch,
	'Rate_Epoch', \@rate_epoch,
	'Unit', \@unit,
	'Constraint', \@constr,
	'ParamVal', \@prmval,
	'ParamSTD', \@stdev,
	'Nparam', $icnt );
    $pkg->{SolutionEstimate_ext} = \%SolEst_ext;

    # my %SolEst= (
    # 	'ParamIndex', \@prmindx,
    # 	'ParamType', \@prmtype,
    # 	'SiteCode', \@sitecode,
    # 	'PointCode', \@pntcode,
    # 	'SolutionID', \@SolID,
    # 	'Epoch', \@epoch,
    # 	'Unit', \@unit,
    # 	'Constraint', \@constr,
    # 	'ParamVal', \@prmval,
    # 	'ParamSTD', \@stdev,
    # 	'Nparam', $icnt );
    # $pkg->{SolutionEstimate} = \%SolEst;

    close(IN);
    return $icnt;

}

# Return solution IDs
# argument1: sitecode
sub SolEstExt_Index( $$ ){
    my $pkg = shift;
    my $sitecode = shift;  ## Site code
    my $sol_est_ext = $pkg->{SolutionEstimate_ext};
    
    my $nparam = $sol_est_ext->{Nparam};
    my $stcdarr =  $sol_est_ext->{SiteCode};
    my $indxs =  $sol_est_ext->{ParamIndex};

    my @rets= ();
    for(my $i=0; $i<$nparam; $i++){
	next unless( $$stcdarr[$i] == $sitecode );
	push(@rets, ($$indxs[$i]));
    }

    return @rets;
}

# Return SolutionID
# argument1: sitecode
# argument1: Epoch
sub Find_SolID( $$$ ){
    my $pkg = shift;
    my $sitecode = shift;  ## SiteCode
    my $epoch = shift;   ## Epoch
    my $sol_est_ext = $pkg->{SolutionEstimate_ext};

    my @idxs = $pkg->SolEstExt_Index( $sitecode );
    foreach $idx (@idxs){
	my $entries = $sol_est_ext->{Entry};
	my $entry = $entries->{$idx};
	my ($t1, $flg1) = &Ext_T1Epoch($entry);
	my ($t2, $flg2) = &Ext_T2Epoch($entry);
	if(  $flg1 eq "NA"){
	    if($flg2 eq "NA"){ next; }
	    elsif( $t2 > $epoch){ 
		return &Ext_Entry_SolID( $entry ); }
	    else{next;}
	}elsif( $t1 > $epoch){next;}
	elsif( $t1 <= $epoch){
	    if( ($flg2 eq "NA") or ($t2 > $epoch)){ return &Ext_Entry_SolID( $entry ); }
	    else{next;}
	}
    }

    return -1;
}

# Return Parameter Index
# argument1: sitecode
# argument1: Epoch
sub Find_ParamIndx( $$$ ){
    my $pkg = shift;
    my $sitecode = shift;  ## SiteCode
    my $epoch = shift;   ## Epoch
    my $sol_est_ext = $pkg->{SolutionEstimate_ext};

    my @idxs = $pkg->SolEstExt_Index( $sitecode );
    my @ret=();
    foreach $idx (@idxs){
	my $entries = $sol_est_ext->{Entry};
	my $entry = $entries->{$idx};
	my ($t1,$flg1) = &Ext_T1Epoch($entry);
	my ($t2,$flg2) = &Ext_T2Epoch($entry);
	if(  $flg1 eq "NA"){
	    if($flg2 eq "NA"){ next; }
	    elsif( $t2 > $epoch){ push(@ret, ($idx) ); next;}
	    else{next;}
	}elsif( $t1 > $epoch){next;}
	elsif( $t1 <= $epoch){
	    if( ($flg2 eq "NA") or ($t2 > $epoch)){ push(@ret, ($idx) ); next; }
	    else{next;}
	}
    }

    return @ret;
}


# Return Entry
# index1: ParamIndex 
sub SolEstExt_Entry ( $$ ){
    my $pkg = shift;
    my $pindex = shift;  ## Index
    my $sol_est_ext = $pkg->{SolutionEstimate_ext};
    my $entries =  $sol_est_ext->{Entry};
     return $entries->{$pindex};
}

sub Ext_Entry_SolID($){
    my $entry = shift;
    return &Ext_Entry($entry,6);
}
sub Ext_Entry_Type($){
    my $entry = shift;
    return &Ext_Entry($entry,0);
}


sub Ext_Entry($$){
    my $entry = shift;
    my $idx = shift;
    my @terms = split(/\s+/,$entry );
    
    if($idx >6 and $idx < 11){
	my $t1 = new Time_lib;
	if( $terms[$idx] eq "-"){ return ("" ,"NA");}
	else{ my ($yy,$doy,$sec) = split(/:/,$terms[$idx]);
	      $t1->Set_time_ydhms($yy,$doy,0,0,0);
	      $t1 ->pluseq($sec);
	      return ($t1,"OK");
	}
    }else{
	return $terms[$idx];
    }
}


sub Ext_T1Epoch($){
    my $entry = shift;
    return &Ext_Entry($entry,8);
}
sub Ext_T2Epoch($){
    my $entry = shift;
    return &Ext_Entry($entry,9);
}
sub Ext_RateEpoch($){
    my $entry = shift;
    return &Ext_Entry($entry,10);
}
sub Ext_RefEpoch($){
    my $entry = shift;
    return &Ext_Entry($entry,7);
}

sub PSD_ext_Entry( $$ ){
     my $entry = shift;
     my $indx = shift;
     my @terms = split( /\s+/, $entry);
     return $terms[$indx];
}

sub PSD_ext_Epoch( $$ ){
    my $pkg = shift;
    my $entry = shift;
    my @terms = split( /\s+/, $entry);
    my ($yy,$doy,$sec) = split(/:/, $terms[4]);
    my $t1 = new Time_lib;
    $t1->Set_time_ydhms($yy,$doy,0,0,0);
    $t1 ->pluseq($sec);
    return ($t1,"OK");
} 
    
1;



