# @Version: $Id: SlvSpool.pm 74 2015-10-26 07:17:15Z vlbi $
#
package SlvSpool;
## Package for reading Spool file of SOLVE.

BEGIN{

}

use Time_lib;

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

sub Get_stn($$$){
    my $this = shift;
    my $in = shift;
    my $stn = shift;
    seek( $in, 0, 0);
    my $epoch = new Time_lib;
    my $mode =0;
    my (@xyz, @xyz_err, @xyz_adj, @enh, @enh_err,); 
    while(<$in>){
	if(/Station positions are for epoch:\s*(\d{4}).(\d+).(\d+)-(\d+):(\d+):(\d+)/){
	    my @ymdhms = ($1,$2,$3,$4,$5,$6);
	    $epoch->Set_time_ymd(@ymdhms);
	    $mode++;
	    next;
	}
	
	if($mode >0){
	    if(/$stn\s+\d+ NOAM  X Comp\s+/){
		my @term = split(/\s+/,$');   # ');
		$xyz[0] = $term[0];
		$xyz_adj[0] = $term[2];
		$xyz_err[0] = $term[4];
		$mode++;
		next;
	    }
	    if(/$stn\s+\d+ \S+  Y Comp\s+/){
		my @term = split(/\s+/,$');   # ');
		$xyz[1] = $term[0];
		$xyz_adj[1] = $term[2];
		$xyz_err[1] = $term[4];
		$mode++;
		next;
	    }
	    if(/$stn\s+\d+ \S+  Z Comp\s+/){
		my @term = split(/\s+/,$');   # ');
		$xyz[2] = $term[0];
		$xyz_adj[2] = $term[2];
		$xyz_err[2] = $term[4];
		$mode++;
		next;
	    }
	    if(/$stn\s+\d+ \s+U Comp\s+/){
		my @term = split(/\s+/,$');   # ');
		$enh[2] = $term[0];
		$enh_err[2] = $term[4];
		$mode++;
		next;
	    }
	    if(/$stn\s+\d+ \s+E Comp\s+/){
		my @term = split(/\s+/,$');   # ');
		$enh[0] = $term[0];
		$enh_err[0] = $term[4];
		$mode++;
		next;
	    }
	    if(/$stn\s+\d+ \s+N Comp\s+/){
		my @term = split(/\s+/,$');   # ');
		$enh[1] = $term[0];
		$enh_err[1] = $term[4];
		$mode++;
		last;
	    }
	    
	}
    }
    
    $this->{epoch} = $epoch;
    $this->{xyz} = \@xyz;
    $this->{xyz_adj} = \@xyz_adj;
    $this->{xyz_err} = \@xyz_err;

    $this->{enh} = \@enh;
    $this->{enh_err} = \@enh_err;
    if($mode < 4){
	$this->{posdatOK} = 0; # Flag for Position data is NG
	print STDERR "E XYZ data was not completely obtained.\n";
	return -1;
    }else{
	$this->{posdatOK} = 1;# Flag for Position data is OK
	return 0;
    }
    
}

sub PosdatOK(){
    my $this = shift;
    $this->{posdatOK};
}


sub Get_base($$$$){
    my $this = shift;
    my $in = shift;
    my $stn1 = shift;
    my $stn2 = shift;
    seek( $in, 0, 0);

    my $mode =0;
    my (@xyz, @xyz_err); 
    my $line;
    my @baseline;
    while($line = <$in>){
	if($line =~/Baseline vector components/){
	    $mode=1;	    next;	}
	if($mode >0){
	    if($line =~/$stn1 \d+       to $stn2  \d+\s+/ ||
	       $line =~/$stn2 \d+       to $stn1  \d+\s+/ ){
		my @term = split(/\s+/,$');  # ');
		$baseline[0] = $term[0];
		$baseline[1] = $term[1];
		$this->{baseline} = \@baseline;
		last;
	    }
	}
    }
    
}

sub Read_Resid($$$$){
    my $this = shift;
    my $in = shift;
    my $stn1 = shift;
    my $stn2 = shift;
    seek( $in, 0, 0);
    
    my $mode =0;
    my (@tarr, @rsdarr, @srcarr, @incflag, @scan );
    my $line;
    my $cnt =0;
    my $msg_old="";
    while($line = <$in>){
	if($line =~/Residuals from/){
	    $mode=1;$line = <$in>;  $line = <$in>; 
	    next;	}
	if($mode >0){
	    if($line =~/([\d]+)\s+([!\@])\s+/){
		my $iscan = $1;
		my $flag = $2;
		unless($' =~m!(\S+)\s+(\S+)\s+(\S+)\s+[\d,.]+[\s\t]+([\d,--+,.]+)\s+ns\s+\d/\s*\d*\s+(\d{4}).(\d\d).(\d\d)-([\d:]+)! ){    # '){
 		    print STDERR "Err: Unexpected data: $line.\n";
		    next;
		}
# 		my @tms = split(/\s+/,$'); # ');
# 		if(@tms < 9){
# 		    print STDERR "Err: Unexpected data: $line.\n";
# 		    next;
# 		}
#		$srcarr[$cnt] = $tms[2];
#		$rsdarr[$cnt] = $tms[4] * 1.e-9;
#		my @hms = split(/[.,\-,:]/,$tms[8]);

		my @stn12 = ($1,$2);
		unless( ( ( $stn12[0] eq $stn1 ) && ($stn12[1] eq $stn2))  || 
			( ( $stn12[1] eq $stn1 ) && ($stn12[0] eq $stn2))  ){
		    my $msg = sprintf "Skip: Station(%s %s) are not Rquested ($stn1,$stn2) \n",
		    $stn12[0],$stn12[1];
		    print STDERR $msg if($msg_old ne $msg);
		    $msg_old = $msg;
		    next;
		}
		$srcarr[$cnt] = $3;
		my $st1 = $1;
		$rsdarr[$cnt] = $4 * 1.e-9;
		my @ymd = ($5, $6, $7);
		my @hms = split(/:/,$8);
		$scan[$cnt] = $iscan;

		my $tm = new Time_lib;
		$tm->Set_time_ymd(@ymd,@hms);
		$tarr[$cnt] = $tm;
		if($flag eq '!' ){ $incflag[$cnt] = 1;}
		else { $incflag[$cnt] = 0;}
		$cnt++;
	    }else{
		last;
	    }
	}
    }

    $this->{delrsd} = \@rsdarr;
    $this->{srcarr} = \@srcarr;
    $this->{rsdtarr} = \@tarr;
    $this->{flag} = \@incflag;
    $this->{iscan} = \@scan;
}

sub Read_Resid2($$$$){
    my $this = shift;
    my $in = shift;
    my $stn1 = shift;
    my $stn2 = shift;
    seek( $in, 0, 0);
    
    my $mode =0;
    my (@tarr, @delarr, @delrsd, @delerr, @ratarr, @raterr, @ratrsd, @srcarr, @incflag, @scan );
    my $line;
    my $cnt =0;
    my $msg_old="";
    my $field = "A14A8A2A7A14A8A2A7A4A4A4A4A15A9";
    while($line = <$in>){
	if($line =~/Residuals from/){
	    $mode=1;$line = <$in>;  $line = <$in>; 
	    next;	}
	if($mode >0){
	    if($line =~/([\d]+>*)\s+/){
		my $tmp = $1;
		my $tmp2 = $'; # ';
		my ($iscan,$flag);
		if($tmp =~/(\d+)>/){ $flag = 1; $iscan=$1;
		}else{		     $flag = 0; $iscan=$tmp;   }
		unless($tmp2 =~m!(\S+)\s*/(\S+)\s+(\S+)\s+(\d{4}.\d\d.\d\d-[\d:\.]+)\s! ){    # '){
 		    print STDERR "Err: Unexpected data: $line.\n";
		    next;
		}
		my @stn12 = ($1,$2);
		$srcarr[$cnt] = $3;
		$tmp = $4;
		$tmp2 = $' ; #';
		unless( ( ( $stn12[0] eq $stn1 ) && ($stn12[1] eq $stn2))  || 
			( ( $stn12[1] eq $stn1 ) && ($stn12[0] eq $stn2))  ){
		    my $msg = sprintf "Skip: Station(%s %s) are not Rquested ($stn1,$stn2) \n",   $stn12[0],$stn12[1];
		    print STDERR $msg if($msg_old ne $msg);
		    $msg_old = $msg;
		    next;
		}
		my @ymdhms = split(/[:.-]/,$tmp);
		my $tm = new Time_lib;
		$tm->Set_time_ymd(@ymdhms);
		$tarr[$cnt] = $tm;
		$scan[$cnt] = $iscan;		
#		my @terms = split(/[\s\t]+/,$tmp2);
		my @terms = unpack($field,$tmp2);
		$delarr[$cnt] = $terms[0] * 1.e-12;			       
		$delrsd[$cnt] = $terms[1] * 1.e-12;			       
		$delerr[$cnt] = $terms[3] * 1.e-12;			       		    
		$ratarr[$cnt] = $terms[4] * 1.e-15;			       
		$ratrsd[$cnt] = $terms[5] * 1.e-15;			       
		$raterr[$cnt] = $terms[7] * 1.e-15;			       		    
		if($terms[2] =~/H/){   $incflag[$cnt] = 0;}
		else{	    $incflag[$cnt] = 1;		} 
		$snr[$cnt] = $terms[13] * 1.e-15;			       		    
		$cnt++;
	    }else{
		last;
	    }
	}
    }
    $this->{srcarr} = \@srcarr;
    $this->{flag} = \@incflag;
    $this->{iscan} = \@scan;
    $this->{srcarr} = \@srcarr;
    $this->{rsdtarr} = \@tarr;

    $this->{delarr} = \@delarr;
    $this->{delrsd} = \@delrsd;
    $this->{delerr} = \@delerr;
    $this->{ratarr} = \@ratarr;
    $this->{ratrsd} = \@ratrsd;
    $this->{raterr} = \@raterr;
    $this->{snr} = \@snr;
    return  $cnt;
}

sub Read_Clk($$$){
    my $this = shift;
    my $in = shift;
    my $stn = shift;
    seek( $in, 0, 0);
    
    my (@tarr, @clkarr,@clkerr, $cl1, $cl2, $cl1err, $cl2err, $tcl1, $tcl2 );
    my $line;
    my $cnt =0;
    my $stname;
    while($line = <$in>){
	if($line =~/(\S+)\s+cl 0\s+/ || $line =~/(\S+)\s+CL 0\s+/){
	    $stname = $1;
	    unless($stn eq $stname){
#		print "Skip:$line ";
		next;
	    }
	    my @tms = split(/\s+/,$'); # ');
	    if(@tms < 8){
		print STDERR "Err: cl0 Unexpected data: $line.\n";
		next;
	    }
	    my @ymd = split(m!/!, $tms[0]);
	    $ymd[0] += $ymd[0] < 100 ? 2000 : 1900; 
	    my @hm = split(/:/, $tms[1]);
	    my $tm = new Time_lib;
	    $tm->Set_time_ymd( @ymd, @hm, 0);
	    $tarr[$cnt] = $tm;
	    $clkarr[$cnt] = $tms[2] * 1.e-9;
	    $clkerr[$cnt] = $tms[6] * 1.e-9;
	    $cnt++;
	}elsif($line =~/(\S+)\s+cl 1\s+/ || $line =~/(\S+)\s+CL 1\s+/){
	    unless($stn eq $stname){
#		print "Skip:$line ";
		next;
	    }
	    my @tms = split(/\s+/,$'); # ');
	    if(@tms < 8){
		print STDERR "Err: cl1 Unexpected data: $line.\n";
		next;
	    }
	    my @ymd = split(m!/!, $tms[0]);
	    my @hm = split(/:/, $tms[1]);
	     $tcl1 = new Time_lib;
	    $tcl1->Set_time_ymd( @ymd, @hm, 0);
	    $tms[3] =~/D([-,\d]+)/;
	    my $ord = $1;
	    $cl1 = $tms[2] * 10**($ord);
	    $cl1err = $tms[6] * 10**($ord);
	}elsif($line =~/(\S+)\s+cl 2\s+/ || $line =~/(\S+)\s+CL 2\s+/){
	    unless($stn eq $stname){
#		print "Skip:$line ";
		next;
	    }
	    my @tms = split(/\s+/,$'); # ');
	    if(@tms < 8){
		print STDERR "Err: cl2 Unexpected data: $line.\n";
		next;
	    }
	    my @ymd = split(m!/!, $tms[0]);
	    my @hm = split(/:/, $tms[1]);
	    $tcl2 = new Time_lib;
	    $tcl2->Set_time_ymd( @ymd, @hm, 0);
	    $tms[3] =~/D([-,\d]+)/;
	    my $ord = $1;
	    $cl2 = $tms[2] * 10**($ord); ## (sec/day)
	    $cl2err = $tms[6] * 10**($ord);
	}
    }
    
    my %clkdat;
    $clkdat{stname} = $stn;
    $clkdat{clkerr} = \@clkerr;
    $clkdat{clkarr} = \@clkarr;
    $clkdat{clkepoch} = \@tarr;
    $clkdat{cl1} = $cl1;
    $clkdat{cl2} = $cl2; ## (sec/day)
    $clkdat{cl1epoch} = $tcl1;
    $clkdat{cl2epoch} = $tcl2;
    $clkdat{cl1err} = $cl1err;
    $clkdat{cl2err} = $cl2err;

    unless(defined($this->{clkdbarr})){
	my @clkdbarr;
	$this->{clkdbarr} = \@clkdbarr;
    }
    my $pclkdbarr = $this->{clkdbarr};
    push(@$pclkdbarr,\%clkdat);

#     $this->{clkerr} = \@clkerr;
#     $this->{clkarr} = \@clkarr;
#     $this->{clkepoch} = \@tarr;
#     $this->{cl1} = $cl1;
#     $this->{cl2} = $cl2; ## (sec/day)
#     $this->{cl1epoch} = $tcl1;
#     $this->{cl2epoch} = $tcl2;
#     $this->{cl1err} = $cl1err;
#     $this->{cl2err} = $cl2err;

}

## mode=0 : MJD
## mode=1 : yyyy/mm/dd-hh:mm:ss 
## mode=2 : yyyy/doy-hh:mm:ss 
sub Epoch($$){
    my $this = shift;
    my $mode = shift;
    my $tm = $this->{epoch};
    unless(defined($tm)){
	print STDERR "W SlvSpool::Epoch(): Position Epoch is not in the data\n";
	return undef;
    }
    if($mode == 0){
	return $tm->Mjd();
    }elsif($mode == 1){
	return $tm->Str_fmt3(1);
    }else{
	return $tm->Str_fmt1(1);
    }
}

sub XYZ($){
    my $this = shift;
    my $xyz = $this->{xyz};
    return @{$xyz};
}

sub XYZ_err($){
    my $this = shift;
    my $xyz_err = $this->{xyz_err};
    return @{$xyz_err};
}

sub XYZ_adj($){
    my $this = shift;
    my $xyz_adj = $this->{xyz_adj};
    return @{$xyz_adj};
}

sub ENH($){
    my $this = shift;
    my $enh = $this->{enh};
    return @{$enh};
}

sub ENH_err($){
    my $this = shift;
    my $enh_err = $this->{enh_err};
    return @{$enh_err};
	    }

sub Baseline($){
    my $this = shift;
    my $base = $this->{baseline};
    unless(defined($base)){
	return (0,0);
    }else{
	return @{$base};
    }
}

sub Read($$$$){
    my $this =shift;
    my $fname =shift;
    my $stn1 =shift;
    my $stn2 =shift;
    my $in;
    unless(open($in,$fname)){
	print STDERR "Cannot opne file $fname \n";
	return -1;
    }
    my $ret = $this->Get_stn($in,$stn2) if($this->Get_stn($in,$stn1) <0);
    
    $this->Get_base($in,$stn1,$stn2) if(defined($stn1) && defined($stn2));
    $this->Read_Resid2($in,$stn1,$stn2) if(defined($stn1) && defined($stn2));
    $this->Read_Clk($in,$stn1) if(defined($stn1));
    $this->Read_Clk($in,$stn2) if(defined($stn2));
    close($in);
    return 0;
}

sub RsdTarr($){
    my $pkg = shift;
    return @{$pkg->{rsdtarr}};
}

sub Srcarr($){
    my $pkg = shift;
    return @{$pkg->{srcarr}};
}

sub Residual($){
    my $pkg = shift;
    return   @{$pkg->{delrsd}};
}

## Solusion included flag array return 
sub Flag($){
    my $pkg = shift;
    return   @{$pkg->{flag}};
}

sub iScan($){
    my $pkg = shift;
    return   @{$pkg->{iscan}};
}

sub ClkEpoch($$){
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:ClkEpoch(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   @{$clkd->{clkepoch}};
	}
    }
    return   ();
}

sub ClkArr($$){
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:ClkArr(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   @{$clkd->{clkarr}};
	}
    }
    return   ();
}

sub ClkErr($){
    my $pkg = shift;
    my $stname = shift;
    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   @{$clkd->{clkerr}};
	}
    }
    return   ();
}

sub Clk1Epoch($$){ # Epoch of clock 1st order
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:Clk1Epoch(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   @{$clkd->{cl1epoch}};
	}
    }
    return   ();
}

sub Clk1($$){ # clock 1st order
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:Clk1(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   $clkd->{cl1};
	}
    }
    return   undef;
}

sub Clk1Err($$){ # clock 1st order error
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:Clk1Err(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   $clkd->{cl1err};
	}
    }
    return   undef;

}

sub Clk2Epoch($$){ # Epoch of clock 2nd order
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:Clk2Epoch(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   $clkd->{cl2epoch};
	}
    }
    return   undef;
}

sub Clk2($$){ # clock 2nd order
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:Clk2(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   $clkd->{cl2};
	}
    }
    return   undef;
}

sub Clk2Err($$){ # clock 2nd order error
    my $pkg = shift;
    my $stname = shift;
    unless(defined($stname)){
	print STDERR "E SlvSpool:Clk2Err(): Stname is not specified.\n";
	return (-1,-1);
    }

    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    return   $clkd->{cl2err};
	}
    }
    return   undef;
}

sub GetClock($$$){ # Compute Clock value with all clock info
    my $pkg = shift;
    my $stname = shift;
    my $xmjd = shift;

    unless(defined($stname)){
	print STDERR "E SlvSpool:GetClock(): Stname is not specified.\n";
	return (-1,-1);
    }
    my (@clkarr,@clkerr,@clkepoch, $cl1, $cl1err, $cl1epoch,$cl2,$cl2err, $cl2epoch);
    my @clkdat =@{$pkg->{clkdbarr}};
    foreach $clkd (@clkdat){
	if($clkd->{stname} eq $stname){
	    @clkarr = @{$clkd->{clkarr}};
	    @clkerr = @{$clkd->{clkerr}};
	    @clkepoch = @{$clkd->{clkepoch}};
	    $cl1 = $clkd->{cl1};
	    $cl1err = $clkd->{cl1err};
	    $cl1epoch = $clkd->{cl1epoch};
	    $cl2 = $clkd->{cl2};
	    $cl2err = $clkd->{cl2err};
	    $cl2epoch = $clkd->{cl2epoch};
	}
    }

    ## Clock params are not estimated.
    return (0,0) if( $#clkarr == -1) ;
    

    my @irange = (0,0);
    my $i=0;
    for($i=0; $i<@clkarr; $i++){
	if($clkepoch[$i]->Mjd() > $xmjd){
	    $irange[0] = $i-1;
	    $irange[1] = $i;
	    last;
	}
    }
    my $ofst;
    if($i >= @clkarr){
	    $irange[0] = $#clkarr - 1;
	    $irange[1] = $#clkarr;
    }
#     $ofst = ($xmjd - $clkepoch[$irange[0]]->Mjd())*86400.*
# 	($clkarr[$irange[1]] -$clkarr[$irange[0]])
# 	/($clkepoch[$irange[1]] - $clkepoch[$irange[0]] ); 
    
    $ofst = 0;
    $ofst += $clkarr[$irange[0]];

    my $err1 = $clkerr[$irange[0]];
    
    my ($err2, $err3) = (0,0);
    my $ret = $ofst;
    if(defined($cl1epoch)){
	my $dt1 = ($xmjd - $cl1epoch->Mjd())*86400.;
	$ret += $cl1 * $dt1; 
	$err2 = $cl1err * $dt1;
#	printf "$stname cl1:$cl1 dt1:$dt1 \n"	;
    }else{
	my $msg = sprintf "Warn: CL 1 absent";
	print STDERR "$msg \n";
    }
    
    if(defined($cl2epoch)){
	my $dt2 = $xmjd - $cl2epoch->Mjd();    
	$ret += $cl2 * $dt2 * $dt2 * 86400.; 
	$err3 = $cl2err * $dt2 * $dt2 * 86400.;
    }else{
	print STDERR "Warn: CL 2 absent \n";
    }
#    my $ret = $ofst;
    my $err = sqrt($err1**2 + $err2**2 + $err3**2);
#    my $err = $err1;
    return ($ret, $err);
    
}

1;
