# @Version: $Id: Time_lib.pm 81 2016-01-21 10:56:01Z vlbi $
#
package VEX;
#use Exporter;
#use Carp;
use Time_lib;

BEGIN{

}

sub new ( $ ){
    my $pkg = shift;
    my $this = {};
    bless $this;
    return $this;
}

sub Read( $$ ){
    my ($pkg, $infile) = @_;

    $pkg->Get_Exp($infile);
    my $nstn = $pkg->Get_Station($infile);
    my $nsite = $pkg->Get_Site($infile);
    my $nsrc = $pkg->Get_Source($infile);
    my $nskd = $pkg->Get_Sched($infile);
    my $nmod = $pkg->Get_Mode($infile);
    my $nproc = $pkg->Get_Proc($infile);

    $pkg->{Nproc} =$nproc;
    $pkg->{Nmode} =$nmod;
    $pkg->{Nstn} =$nstn;
    $pkg->{Nsite} =$nsite;
    $pkg->{Nsrc} =$nsrc;
    $pkg->{Nskd} =$nskd;
}

sub Nmode( $$ ){
    my $pkg = shift;
    $pkg->{Nmode};
}
sub Nskd( $$ ){
    my $pkg = shift;
    $pkg->{Nskd};
}
sub Nsrc( $$ ){
    my $pkg = shift;
    $pkg->{Nsrc};
}
sub Nsite( $$ ){
    my $pkg = shift;
    $pkg->{Nsite};
}
sub Nstn( $$ ){
    my $pkg = shift;
    $pkg->{Nstn};
}

sub Get_Sched( $$ ){
    my ($pkg, $infile) = @_;
    
    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_SCHED(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $rflag=0;
    while( <$in> ){
	if(m!\$SCHED!){
	    $rflag =1; # found SCHED
	    last;
	}
    }
    if($rflag ==0){
	print STDERR "E VEX:Get_SCHED(): SCHED not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <$in>;
    my (@scanNo, %start, %pre, %post, %dur,  %source, %stid,%mode);
    while( $rflag != 2 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $rflag = 2; next; }
	elsif($line =~m!^scan\s+No(\d+)!){  $scanNo[$icnt]=$1;  $rflag = 11;  next; }
	elsif($line =~m!^endscan!){  $rflag = 12;  $icnt++; next; }
	elsif($rflag == 11){
	    if( $line =~/start=(\d+)y(\d+)d(\d+)h(\d+)m(\d+)s/){ 
		my $t1 =new Time_lib; 
		$t1->Set_time_ydhms($1,$2,$3,$4,$5);
		$start{$scanNo[$icnt]} = $t1;
	    }
	    if( $line =~/source=(\S+);/){ $source{$scanNo[$icnt]} =$1;   }
	    if( $line =~/mode=(\S+);/){ $mode{$scanNo[$icnt]} =$1;   }
	    if( $line =~/station\s*=\s*(.+);/){ 
		my @trms = split(/:/,$1);
		$trms[1] =~/\s*(\d+)\s*sec/; $pr1 = $1;
		$trms[2] =~/\s*(\d+)\s*sec/; $dt1 = $1;
		$trms[3] =~/\s*(\d+)\s*sec/; $ps1 = $1;
		my $dt;
		$dt = $dur{$scanNo[$icnt]}  if( defined($dur{$scanNo[$icnt]}) );
		$dt->{$trms[0]} = $dt1; $dur{$scanNo[$icnt]} = $dt;
		my $pr; 
		$pr = $pre{$scanNo[$icnt]} if(defined($pre{$scanNo[$icnt]}));
		$pr->{$trms[0]} = $pr1;	$pre{$scanNo[$icnt]} = $pr;
		my $ps; 
		$ps = $post{$scanNo[$icnt]} if(defined($post{$scanNo[$icnt]}));
		$ps->{$trms[0]} = $ps1;	$post{$scanNo[$icnt]} = $ps;
	    }
	}
    }continue{
	last if( eof($in) ) ;
	$line =<$in>;
    }
    
    my %sched= (
	'ScanNo', \@scanNo,
	'Start', \%start,
	'Source', \%source,
	'Mode', \%mode,
	'Midob', \%dur,
	'Preob', \%pre,
	'Postob', \%post,
	'Nscan', $icnt
	);
    $pkg->{Sched} = \%sched;
    close($in);
    return $icnt;

}

sub ScanSrc( $$ ){
    my $pkg=shift;
    my $iscan = shift;
    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanSrc(): $iscan is not specified.\n";
	return ();
    }

    my $skd = $pkg->{Sched};
    my $src = $skd->{Source};
    my $siscan = sprintf("%04d",$iscan);
    my $src1 = $src->{$siscan};
    return $src1;

}

# Return Scan Number array
sub ScanNo(){
    my $pkg=shift;
    my $skd = $pkg->{Sched};
    my $scanNo = $skd->{ScanNo};
    return @$scanNo;
}

sub ScanStids( $$ ){
    my $pkg=shift;
    my $iscan = shift;
    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanStids(): $iscan is not specified.\n";
	return ();
    }

    my $skd = $pkg->{Sched};
    my $dur = $skd->{Midob};
    my $siscan = sprintf("%04d",$iscan);
    my $dt1 = $dur->{$siscan};
    my @ids = keys(%$dt1);
    return @ids;
}

# 0: Time_lib
# 1: yyyy.ddd.HH:MM:SS
sub ScanMid( $$$;$ ){
    my $pkg=shift;
    my $iscan=shift;
    my $stid=shift;
    my $mode=shift;

    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanMid(): $iscan is not specified.\n";
	return ();
    }

    my $skd = $pkg->{Sched};

    my $siscan = sprintf("%04d",$iscan);
    my $tst = $skd->{Start}->{$siscan};
    # my $dur = $skd->{Midob}->{$siscan}->{$stid};

    # my $st1 = $tst + $dur;
    
    if(defined($mode) && $mode == 1) {
	return $tst->Str_fmt1(2);
    }else{ return $tst; }
}

sub ScanPreob( $$$;$ ){
    my $pkg=shift;
    my $iscan=shift;
    my $stid=shift;
    my $mode=shift;

    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanPreob(): $iscan is not specified.\n";
	return ();
    }

    my $skd = $pkg->{Sched};
    my $mode_db = $pkg->{Mode};
    my $mode_names = $pkg->{Mode_names};
    my $proc_db = $pkg->{Proc};
#    my $dur = $skd->{Midob};
    my $pre = $skd->{Preob};
    my $mode_arr = $skd->{Mode};
    my $siscan = sprintf("%04d",$iscan);
#    my $dt1 = $dur->{$siscan}->{$stid};
    my $pr1 = $pre->{$siscan}->{$stid};
    my $tst = $skd->{Start};
    my $st1 = $tst->{$siscan};
    my $st2 = $st1 + (-$pr1);

    my $mode_name = $mode_arr->{$siscan};
    # my $mode1;
    # for(my $i=0; $i<@$mode_names; $i++ ){
    # 	if( $$mode_names[$i] =~/$mode_name/ ){
    # 	    $mode1 = $mode_db->{ $$mode_names[$i] };
    # 	}
    # }
    # unless( defined($mode1) ){
    # 	printf STDERR "Err: VEX::ScanPreob(): modename:$mode_name not found\n";
    # 	return " Mode1 node defined\n";    
    # }

    my $mode1 = $mode_db->{$mode_name};
    my $prc_name = $mode1->{PROCEDURES};
    my $proc1 = $proc_db->{$prc_name};
    my $pre_all = $proc1->{setup_always};
    $st3=$st2;
    if( $pre_all=~/on\s*:\s*(\d+)\s*sec/){
	$st3 = $st2 + (- $1 );
    }
    
    if( defined($mode) && $mode == 1) {
	return $st3->Str_fmt1(2);
    }else{ return $st3; }
}

sub ScanPostob( $$$;$ ){
    my $pkg=shift;
    my $iscan=shift;
    my $stid=shift;
    my $mode=shift;

    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanPostob(): $iscan is not specified.\n";
	return ();
    }

    my $skd = $pkg->{Sched};
#    my $dur = $skd->{Midob};
    my $siscan = sprintf("%04d",$iscan);
#    my $dt1 = $dur->{$siscan}->{$stid};
#    my $ps1 = $post->{$siscan}->{$stid};
    my $ps1 = $skd->{Postob}->{$siscan}->{$stid};
    my $tst = $skd->{Start}->{$siscan};

    my $st2 = $tst + $ps1;
    
    if(defined($mode) && $mode == 1) {
	return $st2->Str_fmt1(2);
    }else{ return $st2; }
}


sub ScanDur( $$$ ){
    my $pkg=shift;
    my $iscan=shift;
    my $stid=shift;
    
    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanDur(): $iscan is not specified.\n";
	return ();
    }
    my $skd = $pkg->{Sched};
    my $dur = $skd->{Midob};
    my $siscan = sprintf("%04d",$iscan);
    my $dt1 = $dur->{$siscan};
    my @ids = keys(%$dt1);
    
    return $dt1->{$stid};
}

# 0: Time_lib
# 1: yyyy.ddd.HH:MM.SS
sub ScanStop( $$$;$ ){
    my $pkg=shift;
    my $iscan=shift;
    my $stid=shift;
    my $mode=shift;
    
    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanStop(): $iscan is not specified.\n";
	return ();
    }
    my $skd = $pkg->{Sched};
    my $dur = $skd->{Midob};
    my $siscan = sprintf("%04d",$iscan);
    my $dt1 = $dur->{$siscan};
    my @ids = keys(%$dt1);
    my $dt = $dt1->{$stid};
    
    my $start = $skd->{Start};
    my $st1 = $start->{$siscan};
    my $et = $st1 +$dt;

    if(defined($mode) && $mode==1 ){ return $et->Str_fmt1(2);    }
    else{ return $et; }
 
}



sub Get_Site( $$ ){
    my ($pkg, $infile) = @_;
    
    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_Site(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $mode=0;
    while( <$in> ){
	if(m!^\$SITE;!){
	    $mode =1; # found SITE
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E VEX:Get_Site(): SITE not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <$in>;
    my (%sitNo, @sitname, @sitID, %lon, %lat, %height,%xyz,%vxyz,%posEpoch);
    while( !eof($in) && $mode != 2 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $mode = 2; next; }
	elsif($line =~m!^def\s+(\S+);!){   $mode = 11;  next; }
	elsif($line =~m!^enddef!){   $mode = 12;  $icnt++; next; }
	elsif($mode == 11){
	    if( $line =~/site_name\s*=\s*(\S+);/){ $sitname[$icnt] = $1;   }
	    if( $line =~/site_ID\s*=\s+(\S+);/){ $sitID[$icnt] =$1;  $sitNo{$1} = $icnt;  }
	    if( $line =~/elev\s*=\s+(\S+);/){ $height{$sitID[$icnt]} =$1;   }
	    if( $line =~/long\s*=\s+(\S+);/){ $lon{$sitID[$icnt]} =$1;   }
	    if( $line =~/lat\s*=\s+(\S+);/){ $lat{$sitID[$icnt]} =$1;   }
	    if( $line =~/site_position\s*=\s*(.+);/){ 
		my @trms = split(/:/,$1); my @xyz1;
		for(my $i=0; $i<3; $i++){ $trms[$i]=~/([\d,--+,.]+)\s+m/; $xyz1[$i]=$1;}
		$xyz{$sitID[$icnt]} = \@xyz1;
	    }
	    if( $line =~/site_velocity\s*=(.+);/){ 
		my @trms = split(/:/,$1); my @xyz1;
		for(my $i=0; $i<3; $i++){ $trms[$i]=~m!([\d,--+,.]+)\s+m/yr!; $xyz1[$i]=$1;}
		$vxyz{$sitID[$icnt]} = \@xyz1;
	    }
	    if( $line =~/site_position_epoch\s*=(.+);/){ 
		my $tmp = $1;
		unless(defined($posEpoch{$sitID[$icnt]})){
		    my $t1 = new Time_lib;
		    if($tmp=~/(\d+)y(\d+)d/){ 
			$t1->Set_time_ydhms($1,$2,0,0,0);}
		    else{ 
			$t1->Set_time_Mjd($tmp); }
		     $posEpoch{$sitID[$icnt]} = $t1;
		}
	    }
	}
    }continue{
	$line =<$in>;
    }
    
    my %site= (
	'SiteNo', \%sitNo,
	'SiteName', \@sitname,
	'SiteID', \@sitID,
	'Longitude', \%lon,
	'Latitude', \%lat,
	'Height', \%height,
	'XYZ', \%xyz,
	'VXYZ', \%vxyz,
	'PosEpoch', \%posEpoch,
	'Nsite', $icnt
	);
    $pkg->{Site} = \%site;
    close($in);
    return $icnt;

}

# stid -> stname
sub Stname($$){
    my $pkg=shift;
    my $stid=shift;

    my $site = $pkg->{Site};
    my $stname = $site->{SiteName};
    my $sitNo =  $site->{SiteNo};
    my $isit = $sitNo->{$stid};
    unless(defined($isit)){
	print STDERR "E VEX::Stname():stid($stid) is not defined\n";
	my @ids= keys(%{$stNo});
	my $idl = join(" ",@ids);
	print STDERR "Ids are: $idl \n";
	return "";
    }
    return $$stname[$isit];
}

sub Get_Station($$){
    my ($pkg, $infile) = @_;

    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_Station(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $mode=0;
    while( <$in> ){
	if(m!\$STATION!){
	    $mode =1; # found STATION
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E VEX:Get_Station(): STATION not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <$in>;
    my (@stnNo, %site, %antenna, %das);
    while( !eof($in) && $mode != 2 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $mode = 2; next; }
	elsif($line =~m!^def\s+(\S+);!){  $stnNo[$icnt]=$1;  $mode = 11;  next; }
	elsif($line =~m!^enddef!){   $mode = 12;  $icnt++; next; }
	elsif($mode == 11){
	    if( $line =~/ref \$SITE\s*=\s*(\S+);/){ $site{$stnNo[$icnt]} = $1;   }
	    if( $line =~/ref \$ANTENNA\s*=\s*(\S+);/){ $antenna{$stnNo[$icnt]} = $1;   }
	    if( $line =~/ref \$DAS\s*=\s*(\S+);/){ $das{$stnNo[$icnt]} = $1;   }
	}
    }continue{
	$line =<$in>;
    }
    
    my %station= (
	'StationNo', \@stnNo,
	'SiteRef', \%site,
	'AntennaRef', \%antenna,
	'DasRef', \%das,
	'Nstation', $icnt
	);
    $pkg->{Station} = \%station;
    close($in);
    return $icnt;

}

sub Get_Source($$){
    my ($pkg, $infile) = @_;
    
    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_Source(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $mode=0;
    while( <$in> ){
	if(m!\$SOURCE!){
	    $mode =1; # found SOURCE
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E VEX:Get_Source(): SOURCE not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <$in>;
    my (%srcNo, @srcname, @ra, @dec, @ref_frame);
    while( !eof($in) && $mode != 2 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $mode = 2; next; }
	elsif($line =~m!^def\s+(\S+);!){   $mode = 11;  next; }
	elsif($line =~m!^enddef!){
	    $mode = 12;  $icnt++; next; }
	elsif($mode == 11){
	    if( $line =~/source_name\s*=\s*(\S+);/){ 
		$srcname[$icnt] = $1; $srcNo{$1} = $icnt;  }
	    if( $line =~/ra\s*=\s*(\S+);/){ 
		my @trms = split(/[hms]/,$1); $ra[$icnt] = join(":",@trms);
	    }
	    if( $line =~/dec\s*=\s*(\S+);/){ 
		my @trms = split(/[d\'\"]/,$1); $dec[$icnt] = join(":",@trms);
	    }
	    if( $line =~/ref_coord_frame\s*=\s*(\S+);/){  $ref_frame[$icnt] = $1; }
	    next;
	}
    }continue{
	$line =<$in>;
    }
    
    my %source= (
	'SourceNo', \%srcNo,
	'SrcName', \@srcname,
	'Ra', \@ra,
	'Dec', \@dec,
	'RefFrame', \@ref_frame,
	'Nsrc', $icnt
	);
    $pkg->{Source} = \%source;
    close($in);
    return $icnt;

}

## mode 0: dd:mm:ss
## mode 1: ddmmss.xx
sub SrcRa($$){
    my $pkg=shift;
    my $srcname=shift;
    my $mode=shift;

    my $srcdb = $pkg->{Source};
    my $srcNo = $srcdb->{SourceNo};
    my $isrc = $srcNo->{$srcname};
    my $arr = $pkg->{Source}->{Ra};
    my $ra0 = $$arr[$isrc];
    if($mode ==0){ return $ra0;   }
    elsif($mode ==1){
	my @dms = split(/:/,$ra0);
	my $ret = sprintf("%02d%02d%07.4f",@dms);
    }
}

## mode 0: dd:mm:ss
## mode 1: ddmmss.xx
sub SrcDec($$){
    my $pkg=shift;
    my $srcname=shift;
    my $mode=shift;

    my $srcdb = $pkg->{Source};
    my $srcNo = $srcdb->{SourceNo};
    my $isrc = $srcNo->{$srcname};
    my $arr = $pkg->{Source}->{Dec};
    my $dec0 = $$arr[$isrc];
    if($mode ==0){ return $dec0;   }
    elsif($mode ==1){
	my @dms = split(/:/,$dec0);
	my $ret = sprintf("%+03d%02d%07.4f",@dms);
    }
}
## mode 0: dd:mm:ss
## mode 1: ddmmss.xx
sub SrcFrame($$){
    my $pkg=shift;
    my $srcname=shift;

    my $srcdb = $pkg->{Source};
#    my $srcNo = $srcdb->{SourceNo};
#    my $isrc = $srcNo->{$srcname};
    my $isrc = $srcdb->{SourceNo}->{$srcname};
    my $arr = $pkg->{Source}->{RefFrame};
    $$arr[$isrc] =~/\w(\S+)/;
    my $ret = sprintf("%.1f",$1);
    return $ret;  
}


sub Get_Exp($$){
    my ($pkg, $infile) = @_;
    
    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_Exp(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $mode=0;
    while( <$in> ){
	if(m!^\$EXPER!){
	    $mode =1; # found EXP
	    last;
	}
    }
    if($mode ==0){
	print STDERR "E VEX:Get_Exp(): EXPER not found \n";
	return -1;
    }
    
    my $line = <$in>;
    while( !eof($in) && $mode != 2 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $mode = 2; next; }
	elsif($line =~m!^def\s+(\S+);!){  $mode = 11;  next; }
	elsif($line =~m!^enddef!){   $mode = 12;  next; }
	elsif($mode == 11){
	    if( $line =~/exper_name\s*=\s*(\S+);/){ $pkg->{expname} = $1;   }
	    if( $line =~/exper_description\s*=\s*(\S+);/){ $pkg->{exp_descr} = $1;   }
	    if( $line =~/PI_name\s*=\s*(\S+);/){ $pkg->{PI_name} = $1;   }
	    if( $line =~/PI_email\s*=\s*(\S+);/){ $pkg->{PI_email} = $1;   }
	    next;
	}
    }continue{
	$line =<$in>;
    }
    
    close($in);
    return $icnt;

}

sub Expname($){
    my $pkg=shift;
    my $exp = $pkg->{expname};
    return $exp;
}

sub XYZ($$){
    my $pkg = shift;
    my $stid =shift;

    my $site = $pkg->{Site};
    my $xyz = $site->{XYZ};
    unless(defined($xyz->{$stid})){
	print STDERR "E VEX::XYZ(): stid($stid) missing.\n";
	return (0,0,0);
    }
    return @{$xyz->{$stid}};
}

# heading of Snap file
sub Snap_head($$$){
    my $pkg = shift;
    my $stid = shift;
    my $out = shift;
    
    my $stname = $pkg->Stname($stid);
    if( $stname eq "" ){
	print STDERR "E VEX::Snap_head(): Missint Stname\n";
	return -1;
    }
    my @xyz = $pkg->XYZ($stid);
    unless(defined($stname)){
	print STDERR "E VEX::Snap_head(): stid($stid) does not in vex file.\n"; 
	return undef;
    }
    printf $out "\" %s     2016 %s D %s \n", $pkg->Expname(),$stname, $stid;
    printf $out "\" %s %s %+13.5f %+13.5f %+13.5f \n", $stid, $stname, $xyz[0],$xyz[1],$xyz[2];

    return 0;
}

sub Snap_ScanOut($$$$){
    my $pkg=shift;
    my $iscan = shift;
    my $stid = shift;
    my $out = shift;
    my $exp = $pkg->Expname();

    unless(defined($iscan) && $iscan > 0){
	print STDERR "E VEX::ScanStids(): $iscan is not specified.\n";
	return -1;
    }
    my @stids = $pkg->ScanStids($iscan);
    return unless( grep(/$stid/,@stids) );
    printf $out "scan_name=No%04d,%s,%s\n", $iscan, $exp, $stid;
    my $srcname = $pkg->ScanSrc($iscan);
    my $ra = $pkg->SrcRa($srcname,1);
    my $dec = $pkg->SrcDec($srcname,1);
    my $src_epoch = $pkg->SrcFrame($srcname,1);
    printf $out "source=%s,%s,%s,%s\n", $srcname, $ra,$dec,$src_epoch;
    printf $out "!%s\npreob\n",$pkg->ScanPreob($iscan,$stid,1); 
    printf $out "!%s\nmidob\n",$pkg->ScanMid($iscan,$stid,1); 
    printf $out "!%s\npostob\n",$pkg->ScanPostob($iscan,$stid,1); 
    return 0;
}

sub Snap_End($$){
    my $pkg=shift;
    my $out = shift;
    print $out "endproc\n";
    print $out "sched_end\n";

}

sub Get_Mode($$){
    my ($pkg, $infile) = @_;

    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_Mode(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $flag=0;
    while( <$in> ){
	if(m!\$MODE!){
	    $flag =1; # found Mode
	    last;
	}
    }
    if($flag ==0){
	print STDERR "E VEX:Get_Mode(): MODE Entry not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <$in>;
    my (@mode_name, %mode_db);
    while( !eof($in) && $flag != 3 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $flag = 3; next; }
	elsif($line =~m!^def\s+(\S+);!){  
	    $mode_name[$icnt]=$1;  $flag = 21;  
	    my %mod_dat;
	    $mode_db{$mode_name[$icnt]} = \%mod_dat;
	    next; }
	elsif($line =~m!^enddef!){   $flag = 22;  $icnt++; next; }
	elsif($flag == 21){ # inside 'def'
	    if( $line =~/ref\s+\$(\S+)\s*=\s*(\S+);/){ 
		$mode_db{$mode_name[$icnt]}->{$1} = $2;
	    }
	}
    }continue{
	$line =<$in>;
    }
    
    $pkg->{Mode} = \%mode_db;
    $pkg->{Mode_names} = \@mode_name;
    close($in);
    return $icnt;
}

sub Get_Proc($$){
    my ($pkg, $infile) = @_;

    my $in;
    unless(open($in,$infile)){
	print "E VEX::Get_Proc(): Failed to open infile($infile) \n";
	return -1;
    }
    
    my $flag=0;
    while( <$in> ){
	if(m!^\$PROCEDURES!){
	    $flag =1; # found PROCEDURE
	    last;
	}
    }
    if($flag ==0){
	print STDERR "E VEX:Get_Proc(): PROCEDURE Entry not found \n";
	return -1;
    }
    
    my $icnt=0;
    my $line = <$in>;
    my (@proc_name, %proc_db);
    while( !eof($in) && $flag != 3 ){
	next if($line=~/^\*/);
	if($line =~m!^\$!){  $flag = 3; next; }
	elsif($line =~m!^def\s+(\S+);!){  
	    $proc_name[$icnt]=$1;  $flag = 21;  
	    my %mod_dat;
	    $proc_db{$proc_name[$icnt]} = \%mod_dat;
	    next; }
	elsif($line =~m!^enddef!){   $flag = 22;  $icnt++; next; }
	elsif($flag == 21){ # inside 'def'
	    if( $line =~/(\S+)\s*=\s*(.*);/){ 
		$proc_db{$proc_name[$icnt]}->{$1} = $2;
	    }
	}
    }continue{
	$line =<$in>;
    }
    
    $pkg->{Proc} = \%proc_db;
    $pkg->{Proc_names} = \@proc_name;
    close($in);
    return $icnt;
}

## Get SiteID from Station name
sub SiteID_Stname($$){
    my $pkg=shift;
    my $stname = shift;
    my $sitedb=$pkg->{Site};
    my $namearr = $sitedb->{SiteName};
    my $idarr = $sitedb->{SiteID};
    my $nsite = @$namearr;
    my $stid;
    for( my $i=0; $i<$nsite; $i++){
	if($$namearr[$i] =~/$stname/){
	    $stid=$$idarr[$i];
	}
    }

    return $stid;
}


1;
