## @Version: $Id: Mk3db_utl.pm 58 2015-06-10 02:51:51Z vlbi $
##
package Mk3db_utl;
use Exporter;
use Fcntl qw(:seek);
@ISA = qw(Exporter);

@EXPORT = qw(Read_mk3dbhd Baselines BaselineIndex Delarr Delerr Ratarr Raterr Snr CorAmp SrcnameTOC2 SrclistTOC1 Tarr Expname Gambg Heldel Helrat Sapdel Shaprat Delrsd Ratrsd );

BEGIN{
    @TOC1_list = qw( 'UT1EPOCH' 'WOBEPOCH' 'SITERECS' 'AXISOFFS' 'STAR2000'
		   "FUT1 INF" 'FUT1 PTS' 'FWOB INF' 'FWOBX&YT' "TAI- UTC"
		    'SITHOCAM' 'SITOCAMP' 'SITOCPHS' 'SITHOCPH' 'SITNAMES'
                    'STRNAMES' 'END_LIST' );
    @TOC2_list = ('SEC TAG', 'REF FREQ', RFREQ, 'DEL OBSV', 'RAT OBSV', GPDLAMBG,
		    SNRATIO, 'TEMP C', 'ATM PRES',  'REL.HUM.',  'ION CORR', IONRMS, 'STR PART',
		    'ATM PART', NDRYPART, NWETPART, 'EL-THEO', 'AZ-THEO',  'SHAP DEL', 'SHAP RAT',
		    'HELL DEL', 'UTC TAG', RECSETUP, BASELINE, 'STAR ID', 'HELL RAT',
		    CONSNDEL, CONSNRAT,  DELSIGMA, RATSIGMA, DELRESID, RATRESID, COHERCOR, END_LIST);
}

sub new ($){
    my $pkg = shift;
    my $this = {};
    bless $this;
    $this;
}

sub Expname($){
    my $this = shift;
    $this->{expname};
}

sub DBVersion($){
    my $this = shift;
    $this->{db_ver};
}


sub Ratrsd($$){
    my ($this, $ibase )= @_;
    my $ratrsd = $this->{ratrsd};
    $$ratrsd[$ibase];
}

sub Delrsd($$){
    my ($this, $ibase )= @_;
    my $delrsd = $this->{delrsd};
    $$delrsd[$ibase];
}

sub Shaprat($$){
    my ($this, $ibase )= @_;
    my $shaprat = $this->{shaprat};
    $$shaprat[$ibase];
}

sub Shapdel($$){
    my ($this, $ibase )= @_;
    my $shapdel = $this->{shapdel};
    $$shapdel[$ibase];
}

sub Helrat($$){
    my ($this, $ibase )= @_;
    my $helrat = $this->{helrat};
    $$helrat[$ibase];
}

sub Heldel($$){
    my ($this, $ibase )= @_;
    my $heldel = $this->{heldel};
    $$heldel[$ibase];
}

sub Gambg($$){
    my ($this, $ibase )= @_;
    my $gambg = $this->{gambg};
    $$gambg[$ibase];
}

sub Tarr($$){
    my ($this, $ibase )= @_;
    my $tarr = $this->{tarr};
    $$tarr[$ibase];
}

sub SrclistTOC1($$){
    my ($this, $ibase )= @_;
    my $srclist = $this->{STRNAMES};
    $$srclist[$ibase];
}

sub SrcnameTOC2($$){
    my ($this, $ibase )= @_;
    my $srcarr = $this->{srcname};
    $$srcarr[$ibase];
}

sub CorAmp($$){
    my ($this, $ibase )= @_;
    my $coramp = $this->{coramp};
    $$coramp[$ibase];
}


sub Snr($$){
    my ($this, $ibase )= @_;
    my $snr = $this->{snr};
    $$snr[$ibase];
}

sub Raterr($$){
    my ($this, $ibase )= @_;
    my $raterr = $this->{raterr};
    $$raterr[$ibase];
}

sub Delerr($$){
    my ($this, $ibase )= @_;
    my $delerr = $this->{delerr};
    $$delerr[$ibase];
}

sub Ratarr($$){
    my ($this, $ibase )= @_;
    my $rat = $this->{rat};
    $$rat[$ibase];
}

sub Delarr($$){
    my ($this, $ibase )= @_;
    my $del = $this->{del};
    $$del[$ibase];
}

sub BaselineIndex($){
    my $this = shift;
    $this->{baselines};
}

sub Baselines($){
    my $this = shift;
    my $baselines = $this->{baselines};
    return(keys(%$baselines));
}

sub Stnames($){
    my $this = shift;
    my $stnames = $this->{SITNAMES};
    return(@$stnames);
}

sub Stnxyz($){
    my $this = shift;
    my $starr = $this->{SITERECS};
    return(@$starr);
}


sub Read_mk3dbhd($$)
{
    my ($this,$infile) =@_;
    
    open(IN, $infile) || die "Cannot open Mk3 dbhdfile $infile\n";
    $Dexp = "\[\\d,.,--+,D,e\]+";

    my $line = <IN>;

    while(!eof(IN)){
	if($line =~/base\s+\$(\S+)\s+Version\s+(\d+)\s*Occurrence.*record type[\s\t]+(\d+)/){
	    my $expname = $1;
	    my $db_ver = $2;
	    my $itoc = $3;
	    if($itoc ==1){
		$this->{expname} = $expname;
		$this->{db_ver} = $db_ver;
#		my @terms = split(/\s+/);
		$line = $this->toc1_read(IN);
	    }elsif($itoc ==2){
		$line = $this->toc2_read(IN);
	    }

	}
	$line = <IN>;
    }
    close(IN);
}

sub sel_stn(\@){
    my $stnames = shift;
    my $i=0;
    my $nstn = @$stnames;
    do {
	print STDERR "Please select stations number for baseline (ex. 1 2): \n";
	for($i=0; $i<$nstn; $i++){
	    printf STDERR " %2d: %s \n", $i+1, $$stnames[$i]; 
	}
	my $line = <IN>;
	$line =~/\s*(.*)/;
	$line = $1;
	my @terms = split(/\s+/, $line);
    }while( @terms < 2 || ! defined($$stnames[$terms[0]])
	    || ! defined($$stnames[$terms[1]]));
    return($$stnames[$terms[0]], $$stnames[$terms[1]]);
}

sub toc1_read($$){
    my ($this,$in) = @_;
    my $p = caller;
    $in = $p . "::" . $in;
    
    my $line = <$in>;
    my ($line2, $kword, @ndim);
    my $item =0;
    while($line !~/record type/ && !eof($in)){
	
	if( $line =~/Lcode\s=\s+(\S+.*\S+)\s+Dimensions\s+=\s+\(\s*(.+)\)/){
	    $kword = $1;
	    @ndim = split(/\s+/,$2);
	    $item ++;
	    $ndim[0]=1 if($kword =~/NAMES/);
	    $line = " ";
	}elsif(defined($kword)){
	    my  ($i,$j,$k)=(0,0,0);
	    my @datarr;
	    my @arr1;
	    for($i=0; $i<$ndim[2]; $i++){
		my @arr2;
		for($j=0; $j<$ndim[1]; $j++){
		    my @arr3;
		    for($k=0; $k<$ndim[0]; $k++){
			if ($#datarr < 0 ){
			    do{ $line = <$in>; } until( $line =~/\S+/);
#			    until( $line =~/\S+/){ $line = <$in>; } 
			    $line =~/\s*/; 
			    @datarr = split(/\s+/,$'); # ');
			}
			$arr3[$k] = shift(@datarr);
			$arr3[$k] =~s/D\+/e\+/;
			$arr3[$k] =~s/D\-/e\-/;
		    }
		    $arr2[$j] = \@arr3;
		}
		$arr1[$i] = \@arr2;
	    } ## end for
	    $this->{$kword} = \@arr1;
	    undef($kword);
	    do{ $line = <$in>; } until( $line =~/\S+/);
	}else{ ## 
	    print STDERR "unxexpected line: $line \n";
	    do{ $line = <$in>; } until( $line =~/\S+/);	    
	}
    } ## end while
    return($item);
}

sub toc2_read($$){
    my ($this,$in) = @_;
    my $p = caller;
    $in = $p . "::" . $in;
    seek($in, 0, SEEK_SET);
    my $line = <$in>;
    $itoc =0;
    do{
	if($line =~/record type[\s\t]+(\d+)/){
	    $itoc=$1;
	}
	if ($itoc !=2){
	    do{ $line = <$in>; } until( $line =~/\S+/);
	}
    }while($itoc!=2);
    my @oc_arr;
    my $iocrr=0;
    
    my $idat =0;
    my $kword;
    while(!eof($in)){
	if($line =~/Occurrence\s+(\d+)\s*of record type[\s\t]+(\d+)/){
	    $iocrr = $1;
	    $idat++;
	    my %ocrr_db;
	    $ocrr_db{occurence} = $iocrr;
	    do{ $line = <$in>; } until( $line =~/\S+/);
	    while(!eof($in)){
		if( $line =~/Lcode =\s+(\S+.*\S+)\s+Dimensions\s+=\s+\(\s*(.+)\)/){
		    $kword = $1;
		    @ndim = split(/\s+/,$2);
		    $item ++;
		    if( ($kword =~/NAMES/) || ($kword =~/STAR ID/) ||
			($kword =~/BASELINE/ ) ){
			$ndim[0]=1;
		    }
		    $line = " ";
		}elsif(defined($kword)){
		    my  ($i,$j,$k)=(0,0,0);
		    my @datarr;
		    my @arr1;
		    for($i=0; $i<$ndim[2]; $i++){
			my @arr2;
			for($j=0; $j<$ndim[1]; $j++){
			    my @arr3;
			    for($k=0; $k<$ndim[0]; $k++){
				if ($#datarr < 0 ){
				    do{ $line = <$in>; } until( $line =~/\S+/);
				    $line =~/\s*/; 
				    @datarr = split(/\s+/,$'); # ');
				}
				$arr3[$k] = shift(@datarr);
				$arr3[$k] =~s/D\+/e\+/;
				$arr3[$k] =~s/D\-/e\-/;
			    }
			    $arr2[$j] = \@arr3;
			}
			$arr1[$i] = \@arr2;
		    } ## end for
		    $ocrr_db{$kword} = \@arr1;
		    undef($kword);
		    do{ $line = <$in>; } until( $line =~/\S+/);
		}elsif($line=~/base\s+\S+\s+Version\s+\d+\s*Occurrence/){ 
		    $oc_arr[$idat-1] = \%ocrr_db;
		    last;
		}else{ ## 
		    print STDERR "unxexpected line: $line \n";
		    do{ $line = <$in>; } until( $line =~/\S+/);	    
		}
	    } ## LCODE While loop 
	}else{
	    print STDERR "unxexpected line: $line \n";
	    do{ $line = <$in>; } until( $line =~/\S+/);	    
	}
#	$oc_arr[$idat-1] = \%ocrr_db;
	
    }   ## Occurence Loop 
    $this->{TOC2} = \@oc_arr;
    $this->{n_ocrr} = $idat;
    return($idat);
}


1;


